#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This is meant to exercise ViTables capability to zoom into
multidimensional cells.
It also works to check views of multidimensional attributes.
"""

import tables
import numpy

class Particle(tables.IsDescription):
    """Description of a table record."""
    name = tables.StringCol(16, pos=1)
    lati = tables.IntCol(pos=2)
    vector = tables.Int32Col(shape=(200,), pos=3)
    matrix1 = tables.Int32Col(shape=(2, 200), pos=4)
    matrix2 = tables.FloatCol(shape=(100, 2), pos=5)
    matrix3 = tables.FloatCol(shape=(10, 100, 2), pos=5)
    matrix4 = tables.FloatCol(shape=(2, 10, 100, 2), pos=5)

# Open a file in "w"rite mode
fileh = tables.open_file("MDobjects.h5", mode = "w")
# Create the table with compression 'on' in order to reduce size as
# much as possible
table = fileh.create_table(fileh.root, 'table', Particle, "A table", 
    filters=tables.Filters(complevel=1))
# Append several rows with default values
for i in range(10):
    table.row.append()
table.flush()

# create new arrays
atom1 = tables.IntAtom()
shape1 = (2, 10, 10, 1)
filters1 = tables.Filters(complevel=1)
#(2, 10, 10, 3)
array1 = fileh.create_carray(fileh.root, 'array1', atom1, shape1, 
    filters=filters1)
atom2 = tables.FloatAtom()
shape2 = (2, 10, 10, 3, 1)
filters2 = tables.Filters(complevel=1)
#(2, 10, 10, 3, 200)
array2 = fileh.create_carray(fileh.root, 'array2', atom2, shape2, 
    filters=filters2)

# Add multimensional attributes to the objects
# Integers will go in /table
table.attrs.MD1 = numpy.arange(5, dtype="int8")
table.attrs.MD2 = numpy.arange(10, dtype="int64").reshape(2, 5)

# Complex will go in /array1
array1.attrs.MD1 = numpy.arange(5, dtype="complex128")
array1.attrs.MD2 = numpy.arange(10, dtype="complex128").reshape(2, 5)

# Strings will go in /array2
array2.attrs.MD1 = numpy.array(['Hi', 'world!'], dtype='|S6')
array2.attrs.MD2 = numpy.array([['Hi', 'world!'], 
    ['Hola', 'mon!']], dtype='|S4')

fileh.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
#       This script is based on a script by Ivan Vilata.

"""How to use the filenode module."""

from tables.nodes import filenode
import tables

h5file = tables.open_file('../misc/fnode.h5', 'w')
fnode = filenode.new_node(h5file, where='/', name='fnode_test')

counter = 0
while counter < 100000:
    l = "This is a line inserted programmatically at position {}\n".format(counter)
    fnode.write(l.encode("utf-8"))
    counter += 1

fnode.close()

node = h5file.root.fnode_test
fnode = filenode.open_node(node, 'a+')
fnode.write(bytearray("This is the last line.\n", "utf-8"))

fnode.attrs.content_type = 'text/plain; charset=us-ascii'

fnode.attrs.author = "Ivan Vilata i Balaguer"
fnode.attrs.creation_date = '2004-10-20T13:25:25+0200'
fnode.attrs.keywords_en = ["FileNode", "test", "metadata"]
fnode.attrs.keywords_ca = ["FileNode", "prova", "metadades"]
fnode.attrs.owner = 'ivan'
fnode.attrs.acl = {'ivan': 'rw', '@users': 'r'}

fnode.close()
h5file.close()



#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This is the utilities module. It contains functions that perform
tasks that are required at several parts of the application.
"""

import sys
import os
import traceback
import locale
import re
import logging
import collections
import functools

import numpy
import tables

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
from qtpy.QtWidgets import QFileDialog

import vitables.vtwidgets.renamedlg as renamedlg
from vitables.vtsite import ICONDIR

__docformat__ = 'restructuredtext'

ICONS_DICT = {}
HB_ICONS_DICT = {}
DEFAULT_ENCODING = locale.getdefaultlocale()[1]
log = logging.getLogger(__name__)


def getVTApp():
    """Get a reference to the `VTApp` instance.

    This is useful namely for plugins.
    """

    vtapp = None
    for widget in QtWidgets.qApp.topLevelWidgets():
        if widget.objectName() == 'VTGUI':
            vtapp = widget.vtapp
            break

    return vtapp


def getApp():
    """Return getVTApp."""
    return getVTApp()


def getGui():
    """Small wrapper to hide the fact that vtapp object contains gui.

    :return: main window object
    """
    return getApp().gui


def getModel():
    """Small wrapper to hide that vtapp.gui object contains dbs_tree_model.

    :return: the DBs tree model
    """
    return getGui().dbs_tree_model


def getView():
    """Small wrapper to hide that vtapp.gui object contains dbs_tree_view.

    :return: the DBs tree view
    """
    return getGui().dbs_tree_view


def getSelectedIndexes():
    """Return the list of selected indices."""
    return getView().selectedIndexes()


def getSelectedNodes():
    """Return list of selected nodes."""
    return [getModel().nodeFromIndex(index).node
            for index in getSelectedIndexes()]


def long_action(message=None):
    """Decorator that changes the cursor to the wait cursor.

    Used with functions that take some time finish. The provided
    message is displayed in the status bar while the function is
    running (the status bar is cleaned afterwards). If the message is
    not provided then the status bar is not updated.

    :parameter message: string to display in status bar

    """
    def _long_action(f):
        @functools.wraps(f)
        def __long_action(*args, **kwargs):
            status_bar = getGui().statusBar()
            if message is not None:
                status_bar.showMessage(message)
            QtWidgets.QApplication.setOverrideCursor(
                QtGui.QCursor(QtCore.Qt.WaitCursor))
            try:
                res = f(*args, **kwargs)
            finally:
                QtWidgets.QApplication.restoreOverrideCursor()
                if message is not None:
                    status_bar.clearMessage()
            return res
        return __long_action
    return _long_action


def insertInMenu(menu, entries, uid):
    """Insert entries to the given menu before the action named uid.

    The function accept a QAction/QMenu or an iterable. The entries will
    be added before the action whose name is uid.

    :Parameters:

    - `menu`: the menu or context menu being updated
    - `entries`: QAction/Qmenu object or a list of such objects
    - `uid`: indicates the insertion position for the new entries

    :return: None
    """

    if not isinstance(entries, collections.Iterable):
        entries = [entries]

    if isinstance(entries[0], QtWidgets.QAction):
        menu.insertEntry = menu.insertAction
    elif isinstance(entries[0], QtWidgets.QMenu):
        menu.insertEntry = menu.insertMenu

    for item in menu.actions():
        if item.objectName() == uid:
            for a in entries:
                qa = menu.insertEntry(item, a)


def addToMenu(menu, entries):
    """Add entries at the end of the given menu.

    The function accept a QAction/QMenu or an iterable. Entries will be
    preceded with a separator and added at the end of the menu.

    :Parameters:

    - `menu`: the menu or context menu being updated
    - `entries`: QAction/QMenu object or a list of such objects

    :return: None
    """

    if not isinstance(entries, collections.Iterable):
        entries = [entries]

    if isinstance(entries[0], QtWidgets.QAction):
        menu.addEntry = menu.addAction
    elif isinstance(entries[0], QtWidgets.QMenu):
        menu.addEntry = menu.addMenu

    menu.addSeparator()
    for a in entries:
        menu.addEntry(a)


def addActions(target, actions, actions_dict):
    """Add a list of QActions to a menu or a toolbar.

    This is a helper function which make easier to add QActions to a
    menu or a toolbar. Separators and submenus are also handled by this
    method.

    :Parameters:

    - `target`: the menu or toolbar where actions will be added
    - `actions`: a sequence of keywords/None/QMenu used to get actions
                 from a mapping
    - `actions_dict`: a mapping of actions

    """

    for action in actions:
        if action is None:
            target.addSeparator()
        elif isinstance(action, QtWidgets.QMenu):
            target.addMenu(action)
        else:
            target.addAction(actions_dict[action])


def addToLeafContextMenu(actions, enable_function=None):
    """Add entries at the end of the leaf context menu.

    The function accept a QAction/QMenu or an iterable. Entries will
    be preceded with a separator and added at the end of the
    menu. Actions can be enabled or disabled by enable_function which
    is connected to aboutToShow signal of the context menu.

    :parameter actions: QAction/QMenu object or a list of such objects
    :parameter enable_function: is connected to aboutToShow for menu

    :return: None

    """

    context_menu = getGui().leaf_node_cm
    addToMenu(context_menu, actions)
    if enable_function is not None:
        context_menu.aboutToShow.connect(enable_function)


def addToGroupContextMenu(actions):
    """Add entries at the end of the group context menu.

    The function accept a QAction/QMenu or an iterable. Entries will be
    preceded with a separator and added at the end of the menu.

    :parameter actions: QAction/QMenu object or a list of such objects

    :return: None
    """

    context_menu = getGui().group_node_cm
    addToMenu(context_menu, actions)


def getFileSelector(parent, caption, dfilter, filepath='', settings=None):
    """Raise a customised file selector dialog.

    :Parameters:

    - `parent`: the parent widget
    - `caption`: the dialog caption
    - `dfilter`: filters used to display files and folders
    - `filepath`: the filepath initially selected
    - `settings`: dictionary with keys `label` (Accept button text),
        `history` (file selector history) , `accept_mode` and `file_mode`
    """

    file_selector = QtWidgets.QFileDialog(parent, caption, '', dfilter)
    # Misc. setup
    file_selector.setDirectory(settings['history'][-1])
    file_selector.setAcceptMode(settings['accept_mode'])
    if settings['accept_mode'] == QtWidgets.QFileDialog.AcceptSave:
        file_selector.setOption(QFileDialog.DontConfirmOverwrite)
    file_selector.setFileMode(settings['file_mode'])
    file_selector.setHistory(settings['history'])
    if filepath:
        file_selector.selectFile(filepath)
    if settings['label'] != '':
        file_selector.setLabelText(
            QtWidgets.QFileDialog.Accept, settings['label'])

    # Uncomment next line if you want native dialogs. Removing the comment
    # comes at the price of an annoying KDE warning in your console. See the
    # thread "A dire warning message" (July, 2011) in the pyQt4 mailing list
    # for details.
    # file_selector.setOption(QtWidgets.QFileDialog.DontUseNativeDialog)
    return file_selector


def getFilepath(parent, caption, dfilter, filepath='', settings=None):
    """Raise a file selector dialog and get a filepath.

    :Parameters:

    - `parent`: the parent widget
    - `caption`: the dialog caption
    - `dfilter`: filters used to display files and folders
    - `filepath`: the filepath initially selected
    - `settings`: dictionary with keys `label` (Accept button text),
        `history` (file selector history) , `accept_mode` and `file_mode`
    """

    working_dir = None
    file_selector = \
        getFileSelector(parent, caption, dfilter, filepath, settings)

    # Execute the dialog
    try:
        if file_selector.exec_():  # OK clicked
            filepath = file_selector.selectedFiles()[0]
            # Make sure filepath uses Unix-like separators
            filepath = forwardPath(filepath)
            # Update the working directory
            working_dir = file_selector.directory().canonicalPath()
        else:  # Cancel clicked
            filepath = working_dir = ''
    finally:
        del file_selector
    return filepath, working_dir


def checkFileExtension(filepath):
    """
    Check the filename extension of a given file.

    If the filename has no extension this method adds `.h5`
    extension to it. This is useful when a file is being created or
    saved.

    :Parameter filepath: the full path of the file.

    :Returns: the filepath with the proper extension.
    """

    if not re.search('\.(.+)$', os.path.basename(filepath)):
        ext = '.h5'
        filepath = filepath + ext
    return filepath

#
# Icons related functions
#


def createIcons(large_icons, small_icons, icons_dict):
    """
    Create icons for different components of the GUI.

    The method creates sets of icons for the popup menus and the
    toolbar. It also creates icons for the tree of databases view
    and for the windows displaying leaves.

    :Parameters:

    - `large_icons`: the list of names of icons with size 22x22
    - `small_icons`: the list of names of icons with size 16x16
    - `icons_dict`: the icons dictionary to be updated
    """

    all_icons = large_icons.union(small_icons)

    for name in all_icons:
        icon = QtGui.QIcon()
        if name in large_icons:
            pixmap = QtGui.QPixmap(
                os.path.join(ICONDIR, '22x22', '{0}.png').format(name))
            pixmap.scaled(QtCore.QSize(22, 22), QtCore.Qt.KeepAspectRatio)
            icon.addPixmap(pixmap, QtGui.QIcon.Normal, QtGui.QIcon.On)
        if name in small_icons:
            pixmap = QtGui.QPixmap(
                os.path.join(ICONDIR, '16x16', '{0}.png').format(name))
            icon.addPixmap(pixmap, QtGui.QIcon.Normal, QtGui.QIcon.On)
        icons_dict[name] = icon

    # Add an empty iconSet for the Default button of some dialogs
    icons_dict[''] = QtGui.QIcon()

    # Application icon
    icons_dict['vitables_wm'] = QtGui.QIcon(
        os.path.join(ICONDIR, 'vitables_wm.png'))


def getIcons():
    """Return the icons dictionary to be used by the main window."""

    if not ICONS_DICT:
        large_icons = frozenset([
            # Icons for toolbars
            'document-close', 'document-new', 'document-open',
            'document-save-as', 'application-exit', 'folder-new',
            'edit-copy', 'edit-cut', 'edit-delete', 'edit-paste',
            'help-contents',
            'view-filter', 'delete_filters',
            # Icons for tree pane items
            'file_ro', 'file_rw', 'dbfilters', 'folder',
            'document-open-folder'])

        small_icons = frozenset([
            # Icons for menu items
            'document-close', 'document-new', 'document-open',
            'document-save-as', 'document-open-recent',
            'document-export', 'document-import', 'application-exit',
            'edit-cut', 'edit-copy', 'edit-delete', 'edit-paste',
            'edit-rename',
            'folder-new',
            'view-filter', 'delete_filters',
            'configure',
            'help-about', 'help-contents',
            # Icons for tree pane items
            'image-missing', 'object', 'vlstring',
            'array', 'link_array', 'carray', 'link_carray',
            'earray', 'link_earray', 'vlarray', 'link_vlarray',
            'table', 'link_table', 'filenode',
            # Icons for node views
            'zoom-in',
            # Icons for buttons
            'dialog-cancel', 'dialog-ok'])

        createIcons(large_icons, small_icons, ICONS_DICT)

    return ICONS_DICT


def getHBIcons():
    """Return the icons dictionary to be used by the `Help Browser`."""

    if not HB_ICONS_DICT:
        large_icons = frozenset([
            # Icons for toolbar
            'go-first-view', 'go-previous-view', 'go-next-view',
            'view-refresh', 'bookmarks', 'bookmark_add', 'zoom-in', 'zoom-out',
            'edit-clear-history'])

        small_icons = frozenset([
            # Icons for menu items
            'document-open', 'application-exit',
            'zoom-in', 'zoom-out',
            'go-first-view', 'go-previous-view', 'go-next-view',
            'view-refresh',  'bookmarks', 'bookmark_add',
            # Icons for buttons
            'dialog-ok', 'dialog-cancel', 'list-remove'])

        createIcons(large_icons, small_icons, HB_ICONS_DICT)

    return HB_ICONS_DICT


def formatArrayContent(content):
    """
    Nicely format the contents of a view (table widget) cell.

    Used when the cell contains a ``numpy`` array.

    :Parameter content: the ``numpy`` array contained in the view cell
    """

    if isinstance(content, numpy.string_):
        try:
            return content.decode(DEFAULT_ENCODING)
        except UnicodeDecodeError:
            pass

    ret = numpy.array2string(content, separator=',')
    return ret


def formatObjectContent(content):
    """
    Nicely format the contents of a view (table widget) cell.

    Used in `VLArrays` with `object` pseudo atoms.

    Reading a `VLArray` with `object` pseudo atom returns a list of
    Python objects. This method formats that objects as unicode strings
    via str(content).

    :Parameter content: the Python list contained in the view cell
    """

    try:
        text = str(content)
    except UnicodeEncodeError:
        text = content
    return text


def formatStringContent(content):
    """
    Nicely format the contents of a view (table widget) cell.

    Used in `VLArrays` with `vlstring` or `vlunicode` pseudo atoms.
    If the pseudo atom is `vlstring` the method return a encoded string. If
    the pseudo atom is `vlunicode` then a unicode string is returned.

    :Parameter content: the Python list contained in the view cell
    """

    # Content is a encoded string i.e. a sequence of raw bytes
    if isinstance(content, bytes):
        try:
            fcontent = content.decode(encoding=DEFAULT_ENCODING)
            return fcontent
        except UnicodeDecodeError:
            log.error('UnicodeDecodeError exception')
    return content


def formatExceptionInfo(limit=1):
    """
    Format conveniently the catched exceptions.

    Takes the three-element tuple returned by ``sys.exc_info()`` and transforms
    each element into a more convenient form.

    :Parameter limit: the number of stack trace entries to be printed
    """
    log.error('{0}'.format(traceback.format_exc(limit)))

#
# Path related functions
#


def getHomeDir():
    """
    Get the user's home directory.

    How the directory is obtained depends on the platform. The returned
    path is used in QFileDialog calls.
    """

    if sys.platform.startswith('win'):
        home = os.getenv('HOMEDRIVE') + os.getenv('HOMEPATH')
        home = forwardPath(home)
    else:
        home = os.getenv('HOME')
    return home


def forwardPath(path):
    """Replace backslashes with slashes in a given path.

    :Parameter path: the path being transformed
    """

    while path.count(chr(92)):
        path = path.replace(chr(92), '/')
    return path


def questionBox(title='', text='', info='', detail='', buttons_def=''):
    """The value returned by a question message box with customised buttons.

    :Parameters:

    - `title`: the window title
    - `text`: the primary text
    - `info`: additional informative text
    - `detail`: a detailed text
    - `buttons_def`: mapping with buttons definitions

    """

    qmbox = QtWidgets.QMessageBox()
    qmbox.setIcon(QtWidgets.QMessageBox.Question)
    qmbox.setWindowTitle(title)
    qmbox.setText(text)
    if info:
        qmbox.setInformativeText(info)
    if detail:
        qmbox.setDetailedText(detail)
    qmbox.setDefaultButton(QtWidgets.QMessageBox.NoButton)

    buttons = {}
    for name, (text, role) in buttons_def.items():
        buttons[name] = qmbox.addButton(text, role)

    qmbox.exec_()

    # Find out which button has been returned
    for name in buttons.keys():
        if buttons[name] == qmbox.clickedButton():
            return name
    return


def getFinalName(nodename, sibling, pattern, info):
    """Return the node name to be used when editing a node.

    Some node editing operations may raise naming issues because the
    name wanted for the edited node is already being used by a sibling
    node. These operations are:

    - paste
    - drop
    - rename
    - group creation

    If such naming issue arises, a dialog is displayed, and the user is
    required to decide what to do: choose a new name for the node being
    edited, overwrite the node which is currently using the troubled name
    or cancel the node editing.

    :Parameters:

    - `nodename`: the troubled name
    - `sibling`: a sequence with the sibling nodenames of the edited node
    - `pattern`: a regular expression pattern the nodename must match
    - `info`: the information to be displayed in the dialog
    """

    overwrite = False
    # Bad nodename condition
    nodename_in_sibling = nodename in sibling
    # If repeated, ask for a new nodename
    while nodename_in_sibling:
        dialog = renamedlg.RenameDlg(nodename, pattern, info)
        if dialog.exec_():
            nodename = dialog.action['new_name']
            overwrite = dialog.action['overwrite']
            del dialog
            if overwrite is True:
                break
            # Update the bad nodename condition
            nodename_in_sibling = nodename in sibling
        else:
            del dialog
            return None, None
    return nodename, overwrite


def getLicense():
    """The ``ViTables`` license in Rich Text format."""
    # From: http://stackoverflow.com/a/20885799/548792
    import pkg_resources

    resource_package = __name__
    resource_path = 'LICENSE.html'
    license_text = pkg_resources.resource_string(
        resource_package, resource_path)

    return license_text.decode('UTF-8')


def isDataSourceReadable(leaf):
    """Find out if the PyTables leaf can be read or not.
    This is not a complex test. It simply try to read a small chunk
    of data at the beginning of the dataset. If it cannot then the
    dataset is considered unreadable.
    """

    readable = True
    try:
        leaf.read(0, 1)
    except tables.HDF5ExtError as e:
        readable = False
        log.error(
            """Problems reading records. The dataset seems to be """
            """compressed with the {0} library. Check that it is """
            """installed in your system, please.\n{1}""".format(
                leaf.filters.complib, e.message))
    except ValueError as e:
        readable = False
        log.error('Data read error: {}'.format(e.message))
    finally:
        return readable



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Several simple VLArrays."""

import pickle

import tables
import numpy

# Create a VLArray:
fileh = tables.open_file('vlarray_samples.h5', mode='w')

root = fileh.root
vlarray = fileh.create_vlarray(root, 'vlarray1', 
    tables.Int32Atom(), "ragged array of ints", filters=tables.Filters(1))
# Append some (variable length) rows:
vlarray.append(numpy.array([5, 6]))
vlarray.append(numpy.array([5, 6, 7]))
vlarray.append([5, 6, 9, 8])

# Now, do the same with native Python strings.
vlarray2 = fileh.create_vlarray(root, 'vlarray2', 
    tables.StringAtom(itemsize=2), "ragged array of strings", 
    filters=tables.Filters(1))
vlarray2.flavor = 'python'
# Append some (variable length) rows:
vlarray2.append(['5', '66'])
vlarray2.append(['5', '6', '77'])
vlarray2.append(['5', '6', '9', '88'])

# Test with lists of bidimensional vectors
vlarray3 = fileh.create_vlarray(root, 'vlarray3', tables.Int64Atom(shape=(2,)), 
    "Ragged array of vectors")
a = numpy.array([[1, 2], [1, 2]], dtype=numpy.int64)
vlarray3.append(a)
vlarray3.append(numpy.array([[1, 2], [3, 4]], dtype=numpy.int64))
vlarray3.append(numpy.zeros(dtype=numpy.int64, shape=(0, 2)))
vlarray3.append(numpy.array([[5, 6]], dtype=numpy.int64))
# This makes an error (shape)
#vlarray.append(array([[5], [6]], dtype=int64))
# This makes an error (type)
#vlarray.append(array([[5, 6]], dtype=uint64))

# Tests with strings

# Python flavor
vlarray5 = fileh.create_vlarray(root, 'vlarray5', 
    tables.StringAtom(itemsize=3),  "Ragged array of strings")
vlarray5.flavor = "python"
vlarray5.append(["123", "456", "3"])
vlarray5.append(["456", "3"])

# Binary strings
vlarray6 = fileh.create_vlarray(root, 'vlarray6', tables.UInt8Atom(), 
    "pickled bytes")
data = pickle.dumps((["123", "456"], "3"))
vlarray6.append(numpy.ndarray(buffer=data, dtype=numpy.uint8, shape=len(data)))

# The next is a way of doing the same than before
vlarray7 = fileh.create_vlarray(root, 'vlarray7', tables.ObjectAtom(), 
    "pickled object")
vlarray7.append([["123", "456"], "3"])

# Boolean arrays are supported as well
vlarray8 = fileh.create_vlarray(root, 'vlarray8', tables.BoolAtom(), 
    "Boolean atoms")
# The next lines are equivalent...
vlarray8.append([1, 0])
vlarray8.append([1, 0, 3, 0])  # This will be converted to a boolean
# This gives a TypeError
#vlarray.append([1,0,1])

# Variable length strings
vlarray9 = fileh.create_vlarray(root, 'vlarray9', tables.VLStringAtom(), 
    "Variable Length String")
vlarray9.append("asd")
vlarray9.append("aaana")

# Unicode variable length strings
vlarray10 = fileh.create_vlarray(root, 'vlarray10', tables.VLUnicodeAtom(), 
    "Variable Length Unicode String")
vlarray10.append("aaana")
vlarray10.append("")   # The empty string
vlarray10.append("asd")
vlarray10.append("para\u0140lel")

# Still more vlarrays...
N = 100
shape = (3,3)
numpy.random.seed(10)  # For reproductible results
vlarray11 = fileh.create_vlarray(root, 'vlarray11', 
    tables.Float64Atom(shape=shape), "ragged array of arrays")

k = 0
for i in range(N):
    l = []
    for j in range(numpy.random.randint(N)):
        l.append(numpy.random.randn(*shape))
        k += 1
    vlarray11.append(l)

# Close the file.
fileh.close()




#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Plugin that provides sorting algorithms for the DBs tree.

At the moment only two sorting algorithms are supported, alphabetical and
human (a.k.a. natural sorting). In order to obtain the default ViTables
random order of nodes just disable the plugin.

Once the plugin is enabled it works on any file opened after the enabling.
"""

__docformat__ = 'restructuredtext'
__version__ = '1.2'
plugin_name = 'Tree of DBs sorting'
comment = 'Sorts the display of the databases tree'

import logging
import os
import re
import vitables
from vitables.h5db import dbstreemodel
from vitables.h5db import groupnode
from vitables.h5db import leafnode
from vitables.h5db import linknode
from vitables.plugins.dbstreesort.aboutpage import AboutPage

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets
try:
    import configparser
except ImportError:
    import ConfigParser as configparser


translate = QtWidgets.QApplication.translate

log = logging.getLogger(__name__)


def customiseDBsTreeModel():
    """Slot connected to the convenience dbtree_model_created signal.
    """

    # The absolute path of the INI file
    ini_filename = os.path.join(os.path.dirname(__file__),
                                'sorting_algorithm.ini')
    config = configparser.ConfigParser()
    try:
        config.read(ini_filename)
        initial_sorting = config.get('DBsTreeSorting', 'algorithm')
    except (IOError, configparser.ParsingError):
        log.error(
            translate('DBsTreeSort', 'The configuration file of the '
                      'dbs_tree_sort plugin cannot be read.',
                      'DBsTreeSort error message'))
        return

    # The essence of the plugin is pretty simple, just monkeypatch
    # the insertRows() method of the model to get the desired result.
    # TODO how can the nodes be chronologically sorted?
    if initial_sorting == 'human':
        dbstreemodel.DBsTreeModel.insertRows = humanSort
    elif initial_sorting == 'alphabetical':
        dbstreemodel.DBsTreeModel.insertRows = alphabeticalSort
    else:
        log.warning(
            translate('DBsTreeSort', 'Unknown sorting algorithm: {}.',
                      'DBsTreeSort error message').format(initial_sorting))


def alphabeticalSort(self, position=0, count=1, parent=QtCore.QModelIndex()):
    """Sort nodes alphabetically.

    This method is called during nodes population and when files are
    opened/created.

    :Parameters:

     - `position`: the position of the first row being added.
     - `count`: the number of rows being added
     - `parent`: the index of the parent item.

     :Returns: True if the row is added. Otherwise it returns False.
     """

    # Add rows to the model and update its underlaying data store
    self.layoutAboutToBeChanged.emit()
    first = position
    last = position + count - 1
    self.beginInsertRows(parent, first, last)
    node = self.nodeFromIndex(parent)
    # Children are inserted at position 0, so inserted list must
    # be reversed if we want to preserve their original order
    for file_node in sorted(self.fdelta):
        self.root.insertChild(file_node, position)
    for name in sorted(self.gdelta, reverse=True):
        group = groupnode.GroupNode(self, node, name)
        node.insertChild(group, position)
    for name in sorted(self.ldelta, reverse=True):
        leaf = leafnode.LeafNode(self, node, name)
        node.insertChild(leaf, position)
    for name in sorted(self.links_delta, reverse=True):
        link = linknode.LinkNode(self, node, name)
        node.insertChild(link, position)
    self.dataChanged.emit(parent, parent)
    self.endInsertRows()
    self.layoutChanged.emit()

    # Report views about changes in data
    top_left = self.index(first, 0, parent)
    bottom_right = self.index(last, 0, parent)
    self.dataChanged.emit(top_left, bottom_right)

    return True


def alphanum_key(key):
    """ Turn a string into a list of string and number chunks.
        "z23a" -> ["z", 23, "a"]
    """
    def convert(text): return int(text) if text.isdigit() else text

    return [convert(c) for c in re.split(r'(\d+)', key)]


def humanSort(self, position=0, count=1, parent=QtCore.QModelIndex()):
    """Sort nodes using a human sort algorithm.

    This method is called during nodes population and when files are
    opened/created.

    :Parameters:

    - `position`: the position of the first row being added.
    - `count`: the number of rows being added
    - `parent`: the index of the parent item.

    :Returns: True if the row is added. Otherwise it returns False.
    """

    # Add rows to the model and update its underlying data store
    self.layoutAboutToBeChanged.emit()
    first = position
    last = position + count - 1
    self.beginInsertRows(parent, first, last)
    node = self.nodeFromIndex(parent)
    # Children are inserted at position 0, so inserted list must
    # be reversed if we want to preserve their original order
    for file_node in sorted(self.fdelta):
        self.root.insertChild(file_node, position)
    for name in sorted(self.gdelta, reverse=True, key=alphanum_key):
        group = groupnode.GroupNode(self, node, name)
        node.insertChild(group, position)
    for name in sorted(self.ldelta, reverse=True, key=alphanum_key):
        leaf = leafnode.LeafNode(self, node, name)
        node.insertChild(leaf, position)
    for name in sorted(self.links_delta, reverse=True, key=alphanum_key):
        link = linknode.LinkNode(self, node, name)
        node.insertChild(link, position)
    self.dataChanged.emit(parent, parent)
    self.endInsertRows()
    self.layoutChanged.emit()

    # Report views about changes in data
    top_left = self.index(first, 0, parent)
    bottom_right = self.index(last, 0, parent)
    self.dataChanged.emit(top_left, bottom_right)

    return True


class DBsTreeSort(object):
    """Provides convenience methods and functions for sorting the tree of DBs.
    """

    UID = __name__
    NAME = plugin_name
    COMMENT = comment

    def __init__(self):
        """Class constructor.
        """

        self.vtapp = vitables.utils.getVTApp()
        self.vtapp.dbtree_model_created.connect(customiseDBsTreeModel)

    def helpAbout(self, parent):
        """Full description of the plugin.

        This is a convenience method which works as expected by
        :meth:preferences.preferences.Preferences.aboutPluginPage i.e.
        build a page which contains the full description of the plugin
        and, optionally, allows for its configuration.

        :Parameter about_page: the container widget for the page
        """

        # Plugin full description
        desc = {'version': __version__,
                'module_name': os.path.join(os.path.basename(__file__)),
                'folder': os.path.join(os.path.dirname(__file__)),
                'author': 'Vicent Mas <vmas@vitables.org>',
                'about_text': translate(
                    'DBsTreeSortingPage',
                    '<qt><p>Plugin that provides sorting capabilities to '
                    'the tree of DBs.<p>For every open file, nodes (groups and'
                    ' leaves) are sorted when the file is open. '
                    'Nodes added at a later time are not sorted.<p>At the '
                    'moment only two sorting '
                    'algorithms are supported: human (a.k.a. natural sorting) '
                    'and alphabetical.</qt>',
                    'Text of an About plugin message box')}
        self.about_page = AboutPage(desc, parent)

        return self.about_page



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

#
#       This script is based on a set of scripts by Francesc Alted.

"""Several simple Arrays."""

import numpy
import tables

# Open a new empty HDF5 file
fileh = tables.open_file("array_samples.h5", mode = "w")
# Get the root group
root = fileh.root

# Create an Array
a = numpy.array([-1, 2, 4], numpy.int16)
# Save it on the HDF5 file
hdfarray = fileh.create_array(root, 'array_int16', a, "Signed short array")

# Create a scalar Array
a = numpy.array(4, numpy.int16)
# Save it on the HDF5 file
hdfarray = fileh.create_array(root, 'scalar_array', a, 
    "Scalar signed short array")

# Create a 3-d array of floats
a = numpy.arange(120, dtype=numpy.float64).reshape(20, 3, 2)
# Save it on the HDF5 file
hdfarray = fileh.create_array(root, 'array_f3D', a, "3-D float array")

# Create an array of floats
a = numpy.array([1, 2.7182818284590451, 3.141592], numpy.float)
# Save it on the HDF5 file
hdfarray = fileh.create_array(root, 'array_f', a, "Float array")

# Create a large array
#a = reshape(array(range(2**16), "s"), (2,) * 16)
a = numpy.ones((2,) * 8, numpy.int8)

# Save it on the HDF5 file
hdfarray = fileh.create_array(root, 'array_int8', a, "Large array")

# Create a set of arrays and save them on the HDF5 file
basedim = 4
group = root
dtypes = [numpy.int8, numpy.uint8, numpy.int16, numpy.int, numpy.float32, 
    numpy.float]
i = 1
for dtype in dtypes:
    # Create an array of dtype, with incrementally bigger ranges
    a = numpy.ones((basedim,) * i, dtype)
    # Save it on the HDF5 file
    dsetname = 'array_' + a.dtype.char
    hdfarray = fileh.create_array(group, dsetname, a, "Large array")
    # Create a new group
    group = fileh.create_group(root, 'group' + str(i))
    # increment the range for next iteration
    i += 1

# Close the file
fileh.close()




"""Test class for vtconfig.py"""

import sys

import pytest

from qtpy import QtWidgets
# from qtpy import QtCore
from qtpy import QtGui
# from qtpy.QtTest import QTest

import vitables.utils

@pytest.mark.usefixtures('launcher')
class TestLogger(object):
    @pytest.fixture()
    def config(self, launcher):
        cfg = launcher.vtapp_object.config
        yield cfg
        # Tear down code
        cfg.writeValue('Logger/Paper', QtGui.QColor("#ffffff"))
        cfg.writeValue('Logger/Text', QtGui.QColor("#000000"))
        cfg.writeValue('Logger/Font', QtWidgets.qApp.font())
        cfg.writeValue('Workspace/Background',
                       QtGui.QBrush(QtGui.QColor("#ffffff")))
        cfg.writeValue('Look/currentStyle', cfg.default_style)
        launcher.gui.setGeometry(100, 50, 700, 500)
        cfg.writeValue('Geometry/Position', launcher.gui.saveGeometry())
        launcher.gui.logger_dock.setFloating(False)
        launcher.gui.logger_dock.setVisible(True)
        launcher.gui.file_toolbar.setVisible(True)
        cfg.writeValue('Geometry/Layout', launcher.gui.saveState())
        cfg.writeValue('Geometry/HSplitter', launcher.gui.hsplitter.saveState())
        cfg.writeValue('Session/restoreLastSession', False)
        cfg.writeValue('Session/startupWorkingDir', 'home')
        cfg.writeValue('Session/lastWorkingDir', vitables.utils.getHomeDir())

    def test_credentials(self, launcher, config):
        organization = launcher.app.organizationName()
        product = launcher.app.applicationName()
        version = launcher.app.applicationVersion()
        reg_path = 'HKEY_CURRENT_USER\\Software\\{0}\\{1}'.format(product,
                                                                  version)

        if sys.platform.startswith('win'):
            assert config.organizationName() == product
            assert config.applicationName() == version
            assert config.reg_path == reg_path
        elif sys.platform.startswith('darwin'):
            assert config.organizationName() == product
            assert config.applicationName() == version
        else:
            assert config.organizationName() == organization
            assert config.applicationName() == '-'.join((product, version))

    def test_logger(self, config):
        # Background
        bg = QtGui.QColor('#aabbcc')
        config.writeValue('Logger/Paper', bg)
        assert config.loggerPaper() == bg
        # Foreground
        fg = QtGui.QColor('#ccbbaa')
        config.writeValue('Logger/Text', fg)
        assert config.loggerText() == fg
        # Font
        font = QtGui.QFont('Times New Roman')
        config.writeValue('Logger/Font', font)
        assert config.loggerFont() == font

    def test_workspace(self, config):
        bg = QtGui.QBrush(QtGui.QColor('#aabbcc'))
        config.writeValue('Workspace/Background', bg)
        assert config.workspaceBackground() == bg

    def test_appStyle(self, config):
        style = QtWidgets.QStyleFactory.keys()[-1]
        config.writeValue('Look/CurrentStyle', style)
        assert config.readStyle() == style

    def test_windowGeometry(self, launcher, config):
        # Test the main window position and size (without the window frame)
        # Position means x and y coordinates of the top left corner
        # Size means width and height of the window
        launcher.gui.setGeometry(100, 50, 300, 250)
        config.writeValue('Geometry/Position', launcher.gui.saveGeometry())
        launcher.gui.setGeometry(150, 150, 300, 250)
        assert launcher.gui.restoreGeometry(config.windowPosition())
        assert launcher.gui.geometry().x() == 100
        assert launcher.gui.geometry().y() == 50
        assert launcher.gui.width() == 300
        assert launcher.gui.height() == 250

    def test_dockwidgetState(self, launcher, config):
        # Test the state of the main window's dockwidget
        logger_dock = launcher.gui.logger_dock
        logger_dock.setFloating(True)
        logger_dock.setVisible(False)
        config.writeValue('Geometry/Layout', launcher.gui.saveState())
        logger_dock.setFloating(False)
        assert launcher.gui.restoreState(config.windowLayout())
        assert logger_dock.isFloating()

    def test_toolbarsState(self, launcher, config):
        # Test the state (visibility and position) of the main window's toolbars
        # Note: it seems that position is not saved with saveState()
        ftb = launcher.gui.file_toolbar
        ftb.setVisible(False)
        config.writeValue('Geometry/Layout', launcher.gui.saveState())
        ftb.setVisible(True)
        assert launcher.gui.restoreState(config.windowLayout())
        assert not launcher.gui.file_toolbar.isVisible()

    def test_hsplitterState(self, launcher, config):
        launcher.gui.show()
        launcher.gui.setGeometry(100, 550, 700, 500)
        # Test the state (i.e. sizes) of the splitter
        splitter = launcher.gui.hsplitter
        expected_sizes = splitter.sizes()
        config.writeValue('Geometry/HSplitter', splitter.saveState())
        splitter.setSizes([200, 90])
        assert splitter.restoreState(config.hsplitterPosition())
        assert splitter.sizes() == expected_sizes
        launcher.gui.hide()

    def test_restoreLastSession(self, config):
        config.writeValue('Session/restoreLastSession', True)
        assert config.restoreLastSession()
        # None cannot be converted to a boolean value
        config.writeValue('Session/restoreLastSession', None)
        assert not config.restoreLastSession()

    def test_startupWorkingDir(self, config):
        config.writeValue('Session/startupWorkingDir', 'somepath')
        assert config.startupWorkingDir() == 'home'
        config.writeValue('Session/startupWorkingDir', 'last')
        assert config.startupWorkingDir() == 'last'

    def test_lastWorkingDir(self, config):
        config.writeValue('Session/lastWorkingDir', 1)
        assert config.lastWorkingDir() == vitables.utils.getHomeDir()



#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module is the core of the application: it
controls the managers for specific tasks and contains the logic for reacting
to user's input i.e. the slots connected to every menu entry are defined here.
"""

import os
import time
import sys
import logging

import tables

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.utils
import vitables.filenodeutils as fnutils
import vitables.vtsplash
from vitables.vtsite import ICONDIR

from vitables.preferences import vtconfig
import vitables.preferences.pluginsloader as pluginsloader
from vitables.preferences import preferences

import vitables.queries.querymgr as qmgr

import vitables.vtwidgets.nodenamedlg as nodenamedlg
import vitables.vtwidgets.renamedlg as renamedlg

from vitables.docbrowser import helpbrowser

import vitables.vttables.datasheet as datasheet

import vitables.csv.import_csv as importcsv
import vitables.csv.export_csv as exportcsv

import vitables.vtgui as vtgui

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate
log = logging.getLogger(__name__)


def _makePage(content):
    """Create a page for the About ViTables dialog.

    :Parameter content: the text displayed on the page
    """

    widget = QtWidgets.QWidget()
    widget.setLayout(QtWidgets.QVBoxLayout())
    text_edit = QtWidgets.QTextEdit(widget)
    text_edit.setReadOnly(1)
    text_edit.setAcceptRichText(True)
    text_edit.setText(content)
    widget.layout().addWidget(text_edit)

    return widget


class VTApp(QtCore.QObject):
    """
    The application core.

    :Parameters:

    - `mode`: the opening mode for files passed in the command line
    - `h5files`: a list of files to be open at startup
    - `dblist`: a file that contains a list of files to be open at startup
    """
    # Convenience signals for the plugins. Usually new signals are added
    # when a new plugin is added to ViTables. They are the link between
    # the plugins and the core of the program
    leaf_model_created = QtCore.Signal(QtWidgets.QMdiSubWindow,
                                       name="leafModelCreated")
    dbtree_model_created = QtCore.Signal()
    pluginsLoaded = QtCore.Signal()

    def __init__(self, mode='', dblist='', h5files=None, keep_splash=True):
        """
        Initialize the application.

        This method starts the application: makes the GUI, configure the
        app, instantiates managers needed to control the app. and connect
        signals to slots.
        """

        super(VTApp, self).__init__()

        # Make the main window easily accessible for external modules
        self.setObjectName('VTApp')

        self.is_first_opening = True  # for Open file dialogs

        # Show a splash screen
        logo = QtGui.QPixmap(os.path.join(ICONDIR, "vitables_logo.png"))
        splash = vitables.vtsplash.VTSplash(logo)
        splash.show()
        t_i = time.time()

        # Create the GUI. This is done in 3 steps:
        # - create the main window
        # - create the model/view for the tree of databases
        # - setup the main window
        splash.drawMessage(translate('VTApp', 'Creating the GUI...',
                                     'A splash screen message'))
        self.gui = vtgui.VTGUI(self, vtconfig.getVersion())

        # Apply the configuration stored on disk
        splash.drawMessage(translate('VTApp', 'Configuration setup...',
                                     'A splash screen message'))

        # Instantiate a configurator object for the application
        self.config = vtconfig.Config()
        self.config.applyConfiguration(self.config.readConfiguration())

        # Add import/export CSV capabilities
        self.csv_importer = importcsv.ImportCSV()
        self.csv_exporter = exportcsv.ExportToCSV()

        # Load plugins.
        # Some plugins modify existing menus so plugins must be loaded after
        # creating the user interface.
        # Some plugins modify the tree of databases or datasets displaying so
        # plugins must be loaded before opening any file.
        self.plugins_mgr = \
            pluginsloader.PluginsLoader(self.config.enabled_plugins)
        self.plugins_mgr.loadAll()
        self.pluginsLoaded.emit()
        self.dbtree_model_created.emit()

        # The queries manager
        self.queries_mgr = qmgr.QueriesManager()

        # Print the welcome message
        self.gui.logger.write(
            translate('VTApp',
                      """ViTables {0}\nCopyright (c) 2008-2019 Vicent Mas."""
                      """\nAll rights reserved.""",
                      'Application startup message').format(
                          vtconfig.getVersion()))

        # The list of most recently open DBs
        self.number_of_recent_files = 10
        while len(self.config.recent_files) > self.number_of_recent_files:
            del self.config.recent_files[-1]

        # The File Selector History
        self.file_selector_history = []
        if self.config.initial_working_directory != 'last':
            # Reset the last working directory if the Session/startupWorkingDir
            # setting is not 'last'
            self.config.last_working_directory = os.getcwd()
        self.file_selector_history.append(self.config.last_working_directory)

        # List of HelpBrowser instances in memory
        self.doc_browser = None

        # Filenodes mappig
        self.filenodes_map = {}

        # Restore last session
        if self.config.restore_last_session:
            splash.drawMessage(translate('VTApp', 'Recovering last session...',
                                         'A splash screen message'))
            self.recoverLastSession()

        # Process the command line
        if h5files:
            splash.drawMessage(translate('VTApp', 'Opening files...',
                                         'A splash screen message'))
            self.processCommandLineArgs(mode=mode, h5files=h5files)
        elif dblist:
            splash.drawMessage(translate('VTApp',
                                         'Opening the list of files...',
                                         'A splash screen message'))
            self.processCommandLineArgs(dblist=dblist)

        # Make sure that the splash screen is shown at least for two seconds
        if keep_splash:
            t_f = time.time()
            while t_f - t_i < 2:
                t_f = time.time()
        splash.finish(self.gui)
        del splash

        # Ensure that QActions have a consistent state
        self.gui.updateActions()

        self.gui.dbs_tree_model.rowsRemoved.connect(self.gui.updateActions)
        self.gui.dbs_tree_model.rowsInserted.connect(self.gui.updateActions)

        self.gui.updateWindowMenu()
    # Databases are automatically opened at startup when:
    #
    #     * application is configured for recovering last session
    #     * ViTables is started from the command line with some args
    #

    def recoverLastSession(self):
        """
        Recover the last session.

        This method will attempt to open those files and leaf views that
        were opened the last time the user closed ``ViTables``.
        The lists of files and leaves is read from the configuration
        file.
        The format is::

            ['mode#@#filepath1#@#nodepath1#@#nodepath2, ...',
            'mode#@#filepath2#@#nodepath1#@#nodepath2, ...', ...]
        """

        # Get the list of open files (temporary database is not included)
        for file_data in self.config.session_files_nodes:
            item = file_data.split('#@#')
            # item looks like [mode, filepath1, nodepath1, nodepath2, ...]
            mode = item.pop(0)
            filepath = item.pop(0)
            filepath = vitables.utils.forwardPath(filepath)
            # Open the database --> add the root group to the tree view.
            if self.gui.dbs_tree_model.openDBDoc(filepath, mode):
                self.gui.dbs_tree_view.setCurrentIndex(
                    self.gui.dbs_tree_model.index(0, 0, QtCore.QModelIndex()))
            db_doc = self.gui.dbs_tree_model.getDBDoc(filepath)
            if db_doc is None:
                continue
            # Update the history file
            self.updateRecentFiles(filepath, mode)

            # For opening a node the groups in the nodepath are expanded
            # left to right letting the lazy population feature to work
            for nodepath in item:  # '/group1/group2/...groupN/leaf'
                # Check if the node still exists because the database
                # could have changed since last ViTables session
                node = db_doc.get_node(nodepath)
                if node is None:
                    continue
                # groups is ['', 'group1', 'group2', ..., 'groupN']
                groups = nodepath.split('/')[:-1]
                # Expands the top level group, i.e., the root group.
                # It happens to be the last root node added to model
                # so its row is 0
                group = self.gui.dbs_tree_model.root.childAtRow(0)
                index = self.gui.dbs_tree_model.index(0, 0,
                                                      QtCore.QModelIndex())
                self.gui.dbs_tree_view.expanded.emit(index)
                groups.pop(0)
                # Expand the rest of groups of the nodepath
                while groups:
                    parent_group = group
                    parent_index = index
                    group = parent_group.findChild(groups[0])
                    row = group.row()
                    index = self.gui.dbs_tree_model.index(row, 0, parent_index)
                    self.gui.dbs_tree_view.expanded.emit(index)
                    groups.pop(0)
                # Finally we open the leaf
                leaf_name = nodepath.split('/')[-1]
                leaf = group.findChild(leaf_name)
                row = leaf.row()
                leaf_index = self.gui.dbs_tree_model.index(row, 0, index)
                self.gui.dbs_tree_view.setCurrentIndex(leaf_index)
                self.nodeOpen(leaf_index)

    def processCommandLineArgs(self, mode='', h5files=None, dblist=''):
        """
        Open files passed in the command line.

        :Parameters:

        - `mode`: the opening mode for files passed in the command line
        - `h5files`: a list of files to be open at startup
        - `dblist`: a file that contains a list of files to be open at startup
        """

        bad_line = translate('VTApp',
                             'Opening failed: wrong {0} in line {1} of {2}',
                             'Bad line format')
        # The database manager opens the files (if any)
        if isinstance(h5files, list):
            for filepath in h5files:
                filepath = vitables.utils.forwardPath(filepath)
                if self.gui.dbs_tree_model.openDBDoc(filepath, mode):
                    self.gui.dbs_tree_view.setCurrentIndex(
                        self.gui.dbs_tree_model.index(0, 0,
                                                      QtCore.QModelIndex()))
                    self.updateRecentFiles(filepath, mode)

        # If a list of files is passed then parse the list and open the files
        if dblist:
            try:
                input_file = open(dblist, 'r')
                lines = [i[:-1].split('#@#') for i in input_file.readlines()]
                input_file.close()
                for line in lines:
                    if line == ['']:
                        continue
                    if len(line) != 2:
                        log.error(bad_line.format('format', line,
                                                  dblist))
                        continue
                    [mode, filepath] = line
                    filepath = vitables.utils.forwardPath(filepath)
                    if mode not in ['r', 'a']:
                        log.error(bad_line.format('mode', line,
                                                  dblist))
                        continue
                    if self.gui.dbs_tree_model.openDBDoc(filepath, mode):
                        self.gui.dbs_tree_view.setCurrentIndex(
                            self.gui.dbs_tree_model.index(
                                0, 0, QtCore.QModelIndex()))
                        self.updateRecentFiles(filepath, mode)

            except IOError:
                log.error(
                    translate('VTApp', 'List of HDF5 files not read',
                              'File not updated error'))

    def updateRecentFiles(self, filepath, mode):
        """
        Add a new path to the list of most recently open files.

        :meth:`processCommandLineArgs`, :meth:`recoverLastSession`,
        :meth:`fileNew` and :meth:`fileOpen` call this method.

        :Parameters:

            - `filepath`: the last opened/created file
            - `mode`: the opening mode of the file
        """

        item = mode + '#@#' + filepath
        recent_files = self.config.recent_files
        # Updates the list of recently open files. Most recent goes first.
        if item not in recent_files:
            recent_files.insert(0, item)
        else:
            recent_files.remove(item)
            recent_files.insert(0, item)
        while len(recent_files) > self.number_of_recent_files:
            del recent_files[-1]

    def updateFSHistory(self, working_dir):
        """Update the navigation history of the file selector widget.

        :Parameter working_dir: the last visited directory
        """

        # Update the Session/lastWorkingDir setting
        self.config.last_working_directory = working_dir
        if working_dir not in self.file_selector_history:
            self.file_selector_history.append(working_dir)
        else:
            while working_dir in self.file_selector_history:
                self.file_selector_history.remove(working_dir)
            self.file_selector_history.append(working_dir)

    def fileNew(self):
        """Create a new file."""

        # Launch the file selector
        fs_args = {'accept_mode': QtWidgets.QFileDialog.AcceptOpen,
                   'file_mode': QtWidgets.QFileDialog.AnyFile,
                   'history': self.file_selector_history,
                   'label': translate('VTApp', 'Create',
                                      'Accept button text for QFileDialog')}
        filepath, working_dir = vitables.utils.getFilepath(
            self.gui,
            translate('VTApp', 'Creating a new file...',
                      'Caption of the File New... dialog'),
            dfilter=translate('VTApp',
                              """HDF5 Files (*.hdf *.h5 *.hd5 *.hdf5);;"""
                              """All Files (*)""",
                              'Filter for the Open New dialog'),
            settings=fs_args)

        if not filepath:
            # The user has canceled the dialog
            return

        # Update the history of the file selector widget
        self.updateFSHistory(working_dir)

        # Check the file extension
        filepath = vitables.utils.checkFileExtension(filepath)

        # Check the returned path
        if os.path.exists(filepath):
            log.error(
                translate('VTApp', 'New file creation failed because file '
                          'already exists.', 'A file creation error'))
            return

        # Open the database and select it in the tree view
        db_doc = self.gui.dbs_tree_model.createDBDoc(filepath)
        if db_doc:
            # The write mode must be replaced by append mode or the file
            # will be created from scratch in the next ViTables session
            self.gui.dbs_tree_view.setCurrentIndex(
                self.gui.dbs_tree_model.index(0, 0, QtCore.QModelIndex()))
            self.updateRecentFiles(filepath, 'a')

    def fileSaveAs(self):
        """
        Save a renamed copy of a file.

        This method exhibits the typical behavior: copied file is closed
        and the fresh renamed copy is opened.
        """

        overwrite = False
        current_index = self.gui.dbs_tree_view.currentIndex()

        # The file being saved
        initial_filepath = \
            self.gui.dbs_tree_model.nodeFromIndex(current_index).filepath

        # Launch the file selector
        fs_args = {'accept_mode': QtWidgets.QFileDialog.AcceptSave,
                   'file_mode': QtWidgets.QFileDialog.AnyFile,
                   'history': self.file_selector_history,
                   'label': translate('VTApp', 'Create',
                                      'Accept button text for QFileDialog')}
        trier_filepath, working_dir = vitables.utils.getFilepath(
            self.gui,
            translate('VTApp', 'Copying a file...',
                      'Caption of the File Save as... dialog'),
            dfilter=translate('VTApp',
                              """HDF5 Files (*.hdf *.h5 *.hd5 *.hdf5);;"""
                              """All Files (*)""",
                              'Filter for the Save As... dialog'),
            filepath=initial_filepath,
            settings=fs_args)

        if not trier_filepath:  # The user has canceled the dialog
            return

        # Update the history of the file selector widget
        self.updateFSHistory(working_dir)

        trier_filepath = vitables.utils.checkFileExtension(trier_filepath)

        #
        # Check if the chosen name is valid
        #

        info = [translate('VTApp', 'File Save as: file already exists',
                          'A dialog caption'), None]
        # Bad filepath conditions
        trier_dirname, trier_filename = os.path.split(trier_filepath)
        sibling = os.listdir(trier_dirname)
        filename_in_sibling = trier_filename in sibling
        is_tmp_filepath = \
            trier_filepath == self.gui.dbs_tree_model.tmp_filepath
        is_initial_filepath = trier_filepath == initial_filepath

        # If the suggested filepath is not valid ask for a new filepath
        # The loop is necessary because the file being saved as and the
        # temporary database can be in the same directory. In this case
        # we must check all error conditions every time a new name is tried
        while is_tmp_filepath or is_initial_filepath or filename_in_sibling:
            if is_tmp_filepath:
                info[1] = translate('VTApp',
                                    """Target directory: {0}\n\nThe Query """
                                    """results database cannot be """
                                    """overwritten.""",
                                    'Overwrite file dialog label').format(
                    trier_dirname)
                template = \
                    "(^{0}$)|[a-zA-Z_]+[0-9a-zA-Z_]*(?:\.[0-9a-zA-Z_]+)?$"
                pattern = template.format(trier_filename)
            elif is_initial_filepath:
                info[1] = translate('VTApp',
                                    """Target directory: {0}\n\nThe file """
                                    """being saved cannot overwrite itself.""",
                                    'Overwrite file dialog label').format(
                    trier_dirname)
                template = \
                    "(^{0}$)|[a-zA-Z_]+[0-9a-zA-Z_]*(?:\.[0-9a-zA-Z_]+)?$"
                pattern = template.format(trier_filename)
            elif filename_in_sibling:
                info[1] = translate('VTApp',
                                    """Target directory: {0}\n\nFile name {1}"""
                                    """ already in use in that directory.\n""",
                                    'Overwrite file dialog label').format(
                    trier_dirname, trier_filename)
                pattern = "[a-zA-Z_]+[0-9a-zA-Z_]*(?:\.[0-9a-zA-Z_]+)?$"

            dialog = renamedlg.RenameDlg(trier_filename, pattern, info)
            if dialog.exec_():
                trier_filename = dialog.action['new_name']
                trier_filepath = os.path.join(trier_dirname, trier_filename)
                trier_filepath = \
                    QtCore.QDir.fromNativeSeparators(trier_filepath)
                overwrite = dialog.action['overwrite']
                # Update the error conditions
                is_initial_filepath = trier_filepath == initial_filepath
                is_tmp_filepath = \
                    trier_filepath == self.gui.dbs_tree_model.tmp_filepath
                filename_in_sibling = trier_filename in sibling
                del dialog
                if (overwrite is True) and (not is_initial_filepath) and \
                        (not is_tmp_filepath):
                    break
            else:
                del dialog
                return

        filepath = vitables.utils.checkFileExtension(trier_filepath)

        # If an open file is overwritten then close it
        if overwrite and self.gui.dbs_tree_model.getDBDoc(filepath):
            for row, child in enumerate(self.gui.dbs_tree_model.root.children):
                if child.filepath == filepath:
                    self.fileClose(self.gui.dbs_tree_model.index(
                        row, 0, QtCore.QModelIndex()))
            # The current index could have changed when overwriting
            # so we update it
            for row in range(0,
                             self.gui.dbs_tree_model.rowCount(
                                 QtCore.QModelIndex())):
                index = QtCore.QModelIndex().child(row, 0)
                node = self.gui.dbs_tree_model.nodeFromIndex(index)
                if node.filepath == initial_filepath:
                    current_index = index
            self.gui.dbs_tree_view.setCurrentIndex(current_index)

        # Make a copy of the selected file
        try:
            QtWidgets.qApp.setOverrideCursor(
                QtGui.QCursor(QtCore.Qt.WaitCursor))
            dbdoc = self.gui.dbs_tree_model.getDBDoc(initial_filepath)
            dbdoc.copy_file(filepath)
        finally:
            QtWidgets.qApp.restoreOverrideCursor()

        # Close the copied file (which is not necessarely selected in
        # the tree view because closing an overwritten file can change
        # the selected item) and open the new copy in read-write mode.
        # The position in the tree is kept
        for row, child in enumerate(self.gui.dbs_tree_model.root.children):
            if child.filepath == initial_filepath:
                self.fileClose(self.gui.dbs_tree_model.index(
                    row, 0, QtCore.QModelIndex()))
                self.fileOpen(filepath, 'a', row)

    def fileOpenRO(self, filepath=None):
        """
        Open a file that contains a ``PyTables`` database in read-only mode.

        :Parameters filepath: the full path of the file to be open
        """
        self.fileOpen(filepath, mode='r')

    def openRecentFile(self):
        """
        Open the file selected in the `Open Recent...` submenu.
        """

        action = self.sender()
        item = action.data()
        (mode, filepath) = item.split('#@#')
        self.fileOpen(filepath, mode)

    def clearRecentFiles(self):
        """
        Clear the list of recently opened files and delete the corresponding
        historical file.
        """
        self.config.recent_files = []

    def fileOpen(self, filepath=None, mode='a', position=0):
        """
        Open a file that contains a ``PyTables`` database.

        If this method is invoqued via ``File -> Open`` then no filepath
        is passed and a dialog is raised. When the method is invoqued
        passing arguments to the command line or via
        :meth:`openRecentFile` or :meth:`fileSaveAs` methods then
        filepath is passed and the dialog is not raised.

        :Parameters:

        - `filepath`: the full path of the file to be open
        - `mode`: the file opening mode. It can be read-write or read-only
        - `position`: position in the tree view of the new file
        """

        if not filepath:
            # Launch the file selector
            fs_args = {'accept_mode': QtWidgets.QFileDialog.AcceptOpen,
                       'file_mode': QtWidgets.QFileDialog.ExistingFile,
                       'history': self.file_selector_history,
                       'label':
                       translate('VTApp', 'Open', 'Accept text for QFileDialog')
                       }
            filepath, working_dir = vitables.utils.getFilepath(
                self.gui,
                translate('VTApp', 'Select a file for opening',
                          'Caption of the File Open... dialog'),
                dfilter=translate('VTApp',
                                  """HDF5 Files (*.hdf *.h5 *.hd5 *.hdf5);;"""
                                  """All Files (*)""",
                                  'Filter for the Open New dialog'),
                settings=fs_args)

            if not filepath:
                # The user has canceled the dialog
                return

            # Update the history of the file selector widget
            self.updateFSHistory(working_dir)

        else:
            # Make sure filepath uses Unix-like separators
            filepath = vitables.utils.forwardPath(filepath)

        # Open the database and select it in the tree view
        if self.gui.dbs_tree_model.openDBDoc(filepath, mode, position):
            self.gui.dbs_tree_view.setCurrentIndex(
                self.gui.dbs_tree_model.index(
                    position, 0, QtCore.QModelIndex()))
            self.updateRecentFiles(filepath, mode)

    def fileClose(self, current=False):
        """
        Close a file.

        First of all this method finds out which database has to be closed.
        Afterwards all views belonging to that database are closed, then
        the object tree is removed from the databases tree and, finally, the
        database is closed.

        :Parameter current:
            the index in the databases tree of a node living in the file being
            closed
        """

        if current is False:
            current = self.gui.dbs_tree_view.currentIndex()
        node = self.gui.dbs_tree_model.nodeFromIndex(current)
        filepath = node.filepath

        # The position of this database in the tree view
        last_row = len(self.gui.dbs_tree_model.root.children) - 1
        for row, child in enumerate(self.gui.dbs_tree_model.root.children):
            if child.filepath == filepath:
                position = row
                break

        # If some leaf of this database has an open view then close it
        for window in self.gui.workspace.subWindowList():
            if window.dbt_leaf.filepath == filepath:
                window.close()

        # The tree model closes the file and delete its root item
        # from the tree view
        self.gui.dbs_tree_model.closeDBDoc(filepath)

        # The root node immediately below the closed node becomes selected
        if position <= last_row:
            index = self.gui.dbs_tree_model.index(position, 0,
                                                  QtCore.QModelIndex())
            self.gui.dbs_tree_view.setCurrentIndex(index)

    def fileCloseAll(self):
        """Close every open file."""

        # The list of top level items to be removed.
        # The temporary database should be closed at quit time only
        open_files = len(self.gui.dbs_tree_model.root.children) - 1
        rows_range = list(range(0, open_files).__reversed__())
        # Reversing is a must because, if we start from 0, row positions
        # change as we delete rows
        for row in rows_range:
            index = self.gui.dbs_tree_model.index(row, 0, QtCore.QModelIndex())
            self.fileClose(index)

    def fileExit(self):
        """
        Safely close the application.

        Saves current configuration on disk, closes opened files and exits.
        """

        # Close all browsers
        if self.doc_browser:
            self.doc_browser.exitBrowser()
        # Save current configuration
        self.config.saveConfiguration()
        # Close every user opened file
        self.fileCloseAll()
        # Close the temporary database
        index = self.gui.dbs_tree_model.index(0, 0, QtCore.QModelIndex())
        self.fileClose(index)

    def tablesNode(self, index):
        """The tables.Leaf instance tied to the given index.

        :Parameter index: the tree of databases model index being retrieved
        """

        # The leaf (data structure) of the databases tree view tied to index
        leaf = self.gui.dbs_tree_model.nodeFromIndex(index)
        # The tables.Leaf instance referenced by that data structure
        pt_node = leaf.node
        if hasattr(pt_node, 'target'):
            return pt_node()
        return pt_node

    @vitables.utils.long_action("Opening node...")
    def nodeOpen(self, current=False):
        """
        Open a leaf node for viewing.

        :Parameter current: the index in the databases tree of the node being
            opened
        """

        if current is False:
            # Open the node currently selected in the tree of databases
            index = self.gui.dbs_tree_view.currentIndex()
        else:
            # When restoring the previous session explicit indexes are passed
            index = current
        leaf = self.tablesNode(index)

        # tables.UnImplemented datasets cannot be read so are not opened
        if isinstance(leaf, tables.UnImplemented):
            QtWidgets.QMessageBox.information(
                self.gui, translate('VTApp', 'About UnImplemented nodes',
                                    'A dialog caption'),
                translate('VTApp',
                          """Actual data for this node are not accesible."""
                          """<br> The combination of datatypes and/or """
                          """dataspaces in this node is not yet supported by """
                          """PyTables.<br>If you want to see this kind of """
                          """dataset implemented in PyTables, please, """
                          """contact the developers.""",
                          'Text of the Unimplemented node dialog'))
            return

        # Leaves that cannot be read are not opened
        if not vitables.utils.isDataSourceReadable(leaf):
            return

        # Track filenodes
        if fnutils.isFilenode(leaf) and (leaf not in self.filenodes_map):
            temp_filenode = fnutils.filenodeToFile(leaf)
            self.filenodes_map[leaf] = (temp_filenode,
                                        fnutils.filenodeTotalRows(leaf))

        # Create a view
        subwindow = datasheet.DataSheet(index)
        subwindow.show()

        # For unknown reasons sometimes views cannot be resized. This strange
        # problem is fixed by hiding the view and immediately showing it again
        subwindow.hide()
        subwindow.show()

        # Announcing the new view is potentially helpful for plugins in charge
        # of datasets customizations (for instance, additional formatting)
        self.leaf_model_created.emit(subwindow)

    def nodeClose(self, current=False):
        """
        Close the view of the selected node.

        The method is called by activating ``Node --> Close`` (what passes
        no argument) or programatically by the :meth:`fileClose`
        method (what does pass argument).
        If the target is an open leaf this method closes its view, delete
        its model and updates the controller tracking system.
        If the target node is a root group the method looks for opened
        children and closes them as described above.

        :Parameter current: the index in the databases tree of the node being
            closed
        """

        if not current:
            current = self.gui.dbs_tree_view.currentIndex()
        pcurrent = QtCore.QPersistentModelIndex(current)
        # Find out the subwindow tied to the selected node and close it
        for data_sheet in self.gui.workspace.subWindowList():
            if pcurrent == data_sheet.pindex:
                data_sheet.close()
                break

    def nodeNewGroup(self):
        """Create a new group node."""

        current = self.gui.dbs_tree_view.currentIndex()
        parent = self.gui.dbs_tree_model.nodeFromIndex(current)

        # Get the new group name
        dialog = nodenamedlg.InputNodeName(
            translate('VTApp', 'Creating a new group', 'A dialog caption'),
            translate('VTApp', 'Source file: {0}\nParent group: {1}\n\n ',
                      'A dialog label').format(parent.filepath,
                                               parent.nodepath),
            translate('VTApp', 'Create', 'A button label'))
        self.gui.editing_dlg = True
        if dialog.exec_():
            suggested_nodename = dialog.node_name
            self.gui.dbs_tree_view.setFocus(True)
            del dialog
        else:
            self.gui.dbs_tree_view.setFocus(True)
            del dialog
            return

        #
        # Check if the entered nodename is already in use
        #
        sibling = getattr(parent.node, '_v_children').keys()
        pattern = "[a-zA-Z_]+[0-9a-zA-Z_ ]*"
        info = [translate('VTApp',
                          'Creating a new group: name already in use',
                          'A dialog caption'),
                translate('VTApp',
                          """Source file: {0}\nParent group: {1}\n\nThere is """
                          """already a node named '{2}' in that parent group."""
                          """\n""",
                          'A dialog label').format
                (parent.filepath, parent.nodepath, suggested_nodename)]
        nodename, overwrite = vitables.utils.getFinalName(suggested_nodename,
                                                          sibling, pattern,
                                                          info)
        if nodename is None:
            return

        # If the creation overwrites a group with attached views then these
        # views are closed before the renaming is done
        if overwrite:
            nodepath = tables.path.join_path(parent.nodepath, nodename)
            self.gui.closeChildrenViews(nodepath, parent.filepath)

        self.gui.dbs_tree_model.create_group(current, nodename, overwrite)

    def nodeRename(self):
        """
        Rename the selected node.

        - ask for the node name
        - check the node name. If it is already in use ask what to
          do (possibilities are rename, overwrite and cancel creation)
        - rename the node
        """

        index = self.gui.dbs_tree_view.currentIndex()
        child = self.gui.dbs_tree_model.nodeFromIndex(index)
        parent = child.parent

        # Get the new nodename
        dialog = nodenamedlg.InputNodeName(
            translate('VTApp', 'Renaming a node', 'A dialog caption'),
            translate('VTApp', 'Source file: {0}\nParent group: {1}\n\n',
                      'A dialog label').format(parent.filepath,
                                               parent.nodepath),
            translate('VTApp', 'Rename', 'A button label'),
            child.name)
        self.gui.editing_dlg = True
        if dialog.exec_():
            suggested_nodename = dialog.node_name
            del dialog
        else:
            del dialog
            return

        #
        # Check if the nodename is already in use
        #
        sibling = getattr(parent.node, '_v_children').keys()
        # Note that current nodename is not allowed as new nodename.
        # Embedding it in the pattern makes unnecessary to pass it to the
        # rename dialog via method argument and simplifies the code
        pattern = """(^{0}$)|""" \
            """(^[a-zA-Z_]+[0-9a-zA-Z_ ]*)""".format(child.name)
        info = [translate('VTApp', 'Renaming a node: name already in use',
                          'A dialog caption'),
                translate('VTApp',
                          """Source file: {0}\nParent group: {1}\n\nThere is """
                          """already a node named '{2}' in that parent """
                          """group.\n""", 'A dialog label').format
                (parent.filepath, parent.nodepath, suggested_nodename)]
        nodename, overwrite = vitables.utils.getFinalName(suggested_nodename,
                                                          sibling, pattern,
                                                          info)
        self.gui.editing_dlg = True
        if nodename is None:
            return

        # If the renaming overwrites a node with attached views then these
        # views are closed before the renaming is done
        if overwrite:
            nodepath = tables.path.join_path(parent.nodepath, nodename)
            self.gui.closeChildrenViews(nodepath, child.filepath)

        # Rename the node
        self.gui.dbs_tree_model.rename_node(index, nodename, overwrite)

        # Update the Selected node indicator of the status bar
        self.gui.updateStatusBar()

    def nodeCut(self):
        """Cut the selected node."""

        current = self.gui.dbs_tree_view.currentIndex()

        # If the cut node has attached views then these views are closed
        # before the cutting is done. This behavior can be inconvenient
        # for users but get rid of potential problems that arise if, for
        # any reason, the user doesn't paste the cut node.
        node = self.gui.dbs_tree_model.nodeFromIndex(current)
        self.gui.closeChildrenViews(node.nodepath, node.filepath)

        # Cut the node
        self.gui.dbs_tree_model.cutNode(current)

    def nodeCopy(self):
        """
        Copy the selected node.
        """

        current = self.gui.dbs_tree_view.currentIndex()

        # Non readable leaves should not be copied
        dbs_tree_node = self.gui.dbs_tree_model.nodeFromIndex(current)
        if not (dbs_tree_node.node_kind in ('root group', 'group')):
            leaf = dbs_tree_node.node
            # tables.Leaf instances must be readable in order to be copied
            if not hasattr(leaf, 'target'):
                if not vitables.utils.isDataSourceReadable(leaf):
                    QtWidgets.QMessageBox.information(
                        self,
                        translate('VTApp', 'About unreadable datasets',
                                  'Dialog caption'),
                        translate('VTApp', """Sorry, actual data for this """
                                           """node are not accesible.<br>The """
                                           """node will not be copied.""",
                                  'Text of the Unimplemented node dialog'))
                    return

        # Copy the node
        self.gui.dbs_tree_model.copy_node(current)

    def nodePaste(self):
        """
        Paste the currently copied/cut node under the selected node.
        """

        current = self.gui.dbs_tree_view.currentIndex()
        parent = self.gui.dbs_tree_model.nodeFromIndex(current)

        cni = self.gui.dbs_tree_model.ccni
        if cni == {}:
            return

        nodename = cni['nodename']
        if cni['nodepath'] == '/':
            nodename = \
                'root_group_of_{0}'.format(os.path.basename(cni['filepath']))

        if cni['is_copied']:
            # Check if pasting a copied node is allowed (pasting a cut
            # node has no restrictions). It is not when
            # - source and target are the same node
            # - target is the source's parent
            if cni['filepath'] == parent.filepath:
                if (cni['nodepath'] == parent.nodepath) or \
                        (parent.nodepath == os.path.dirname(cni['nodepath'])):
                    return

        # Soft links cannot be pasted in external files
        if cni['target']:
            link = self.gui.dbs_tree_model.copiedNode()
            try:
                getattr(link, 'extfile')
            except AttributeError:
                if parent.filepath != cni['filepath']:
                    return

        #
        # Check if the nodename is already in use
        #
        sibling = getattr(parent.node, '_v_children').keys()
        # Nodename pattern
        pattern = "[a-zA-Z_]+[0-9a-zA-Z_ ]*"
        # Bad nodename conditions
        info = [translate('VTApp', 'Node paste: nodename already exists',
                          'A dialog caption'),
                translate('VTApp', """Source file: {0}\nCopied node: {1}\n"""
                          """Destination file: {2}\nParent group: {3}\n\n"""
                          """Node name '{4}' already in use in that group.\n""",
                          'A dialog label').format
                (cni['filepath'], cni['nodepath'], parent.filepath,
                 parent.nodepath, nodename),
                translate('VTApp', 'Paste', 'A button label')]
        # Validate the nodename
        nodename, overwrite = vitables.utils.getFinalName(nodename, sibling,
                                                          pattern, info)
        self.gui.editing_dlg = True
        if nodename is None:
            # Node editing dialog is not displayed
            return

        # If the pasting overwrites a node with attached views then these
        # views are closed before the pasting is done
        if overwrite:
            nodepath = tables.path.join_path(parent.nodepath, nodename)
            self.gui.closeChildrenViews(nodepath, parent.filepath)

        # Paste the node
        self.gui.dbs_tree_model.pasteNode(current, nodename, overwrite)

    def nodeDelete(self, current=False, force=None):
        """
        Delete a given node.

        :Parameters:

            - `current`: the index in the databases tree of the node being
              deleted
            - `force`: ask/do not ask for confirmation before deletion
        """

        if current is False:
            current = self.gui.dbs_tree_view.currentIndex()
        node = self.gui.dbs_tree_model.nodeFromIndex(current)

        # Confirm deletion dialog
        if not force:
            title = translate('VTApp', 'Node deletion',
                              'Caption of the node deletion dialog')
            text = translate('VTApp',
                             """\nYou are about to delete the node:\n{0}\n""",
                             'Ask for confirmation').format(node.nodepath)
            itext = ''
            dtext = ''
            buttons = {
                'Delete':
                    (translate('VTApp', 'Delete', 'Button text'),
                     QtWidgets.QMessageBox.YesRole),
                'Cancel':
                    (translate('VTApp', 'Cancel', 'Button text'),
                     QtWidgets.QMessageBox.NoRole),
            }

            # Ask for confirmation
            answer = \
                vitables.utils.questionBox(title, text, itext, dtext, buttons)
            self.gui.editing_dlg = True
            if answer == 'Cancel':
                return

        # If item is a filtered table then update the list of used names
        if hasattr(node.node._v_attrs, 'query_condition'):
            self.queries_mgr.ft_names.remove(node.name)

        # If the deletion involves a node with attached views then these
        # views are closed before the deletion is done
        self.gui.closeChildrenViews(node.nodepath, node.filepath)

        # Delete the node
        self.gui.dbs_tree_model.deleteNode(current)

        # Ensure that the new current node (if any) gets selected
        select_model = self.gui.dbs_tree_view.selectionModel()
        new_current = self.gui.dbs_tree_view.currentIndex()
        select_model.select(new_current, QtCore.QItemSelectionModel.Select)

    def nodeProperties(self):
        """
        Display the properties dialog for the currently selected node.

        The method is called by activating ``Node --> Properties``.
        """

        current = self.gui.dbs_tree_view.currentIndex()
        node = self.gui.dbs_tree_model.nodeFromIndex(current)
        node.properties()
        self.gui.editing_dlg = True

    def newQuery(self):
        """Slot for querying tables."""
        self.queries_mgr.newQuery()

    def deleteAllQueries(self):
        """Slot for emptying the `Query results` node."""
        self.queries_mgr.deleteAllQueries()

    def settingsPreferences(self):
        """
        Launch the Preferences dialog.

        Clicking the ``OK`` button applies the configuration set in the
        Preferences dialog.
        """

        prefs = preferences.Preferences()
        try:
            if prefs.exec_() == QtWidgets.QDialog.Accepted:
                self.config.applyConfiguration(prefs.new_prefs)
        finally:
            del prefs

    def windowClose(self):
        """Close the window currently active in the workspace."""

        self.gui.workspace.activeSubWindow().close()
        if not self.gui.workspace.subWindowList():
            self.gui.dbs_tree_view.setFocus(True)

    def windowCloseAll(self):
        """Close all open windows."""

        for window in self.gui.workspace.subWindowList():
            window.close()
        self.gui.dbs_tree_view.setFocus(True)

    def windowRestoreAll(self):
        """Restore every window in the workspace to its normal size."""

        for window in self.gui.workspace.subWindowList():
            window.showNormal()

    def windowMinimizeAll(self):
        """Restore every window in the workspace to its normal size."""

        for window in self.gui.workspace.subWindowList():
            window.showMinimized()

    def helpBrowser(self):
        """
        Open the documentation browser.
        """
        self.doc_browser = helpbrowser.HelpBrowser()

    def helpAbout(self):
        """
        Show a tabbed dialog with the application About and License info.
        """

        # Text to be displayed
        about_text = translate('VTApp', """<qt><h3>ViTables {0}</h3>
            ViTables is a graphical tool for displaying datasets
            stored in PyTables and HDF5 files. It is written using PyQt
            , the Python bindings for the Qt GUI toolkit.<p>
            For more information see
            <b>https://vitables.org</b>.<p>
            Please send bug reports or feature requests to the
            <em>ViTables Users Group</em>.<p>
            ViTables uses third party software which is copyrighted by
            its respective copyright holder. For details see the
            copyright notice of the individual packages.
            </qt>""", 'Text of the About ViTables dialog').format(
            vtconfig.getVersion())
        thanks_text = translate('VTApp', """<qt>
            Jens Janssen, Kostis Anagnostopoulos, Arne de Laat and Christoph
            Gohlke for
            helping to revive ViTables with their great contributions.<p>
            Santi Villalba for the conda environment.<p>
            Alexey Naydenov for his contributions to the plugins system and for
            providing logging support to the application.<p>
            Robert Meyer, jaminsore, kprussing, Gaetan de Menten, Kamil Kisiel
            and Max Bohnet for fixing bugs.<p>
            Dmitrijs Ledkovs for contributing the new and greatly enhanced
            build system and for making Debian packages.<p>
            Oxygen team for a wonderful icons set.<p>
            All the people who reported bugs and made suggestions.
            </qt>""", 'Text of the About ViTables dialog (Thanks to page)')
        license_text = vitables.utils.getLicense()

        # Construct the dialog
        about_dlg = QtWidgets.QDialog(self.gui)
        about_dlg.setWindowTitle(
            translate('VTApp', 'About ViTables {0}',
                      'Caption of the About ViTables dialog').
            format(vtconfig.getVersion()))
        layout = QtWidgets.QVBoxLayout(about_dlg)
        tab_widget = QtWidgets.QTabWidget(about_dlg)
        buttons_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
        layout.addWidget(tab_widget)
        layout.addWidget(buttons_box)

        buttons_box.accepted.connect(about_dlg.accept)

        # Make About page
        content = [about_text, thanks_text, license_text]
        tabs = [translate('VTApp', '&About',
                          'Title of the first tab of the About dialog'),
                translate('VTApp', '&Thanks To',
                          'Title of the second tab of the About dialog'),
                translate('VTApp', '&License',
                          'Title of the third tab of the About dialog')]

        for index in range(0, 3):
            widget = _makePage(content[index])
            tab_widget.addTab(widget, tabs[index])

        # Show the dialog
        about_dlg.exec_()

    def helpAboutQt(self):
        """
        Show a message box with the Qt About info.
        """
        QtWidgets.QMessageBox.aboutQt(self.gui, translate(
            'VTApp', 'About Qt', 'Caption of the About Qt dialog'))

    def helpVersions(self):
        """
        Message box with info about versions of libraries used by
        ``ViTables``.
        """

        svi = sys.version_info
        pyversion = '.'.join((str(svi.major), str(svi.minor), str(svi.micro)))
        # The libraries versions dictionary
        libs_versions = {
            'title': translate('VTApp', 'Version Numbers',
                               'Caption of the Versions dialog'),
            'Python': pyversion,
            'PyTables': tables.__version__,
            'NumPy': tables.numpy.__version__,
            'Qt': QtCore.qVersion(),
            'PyQt': QtCore.PYQT_VERSION_STR,
            'ViTables': vtconfig.getVersion()
        }

        # Add new items to the dictionary
        libraries = ('HDF5', 'Zlib', 'LZO', 'BZIP2')
        for lib in libraries:
            lversion = tables.which_lib_version(lib.lower())
            if lversion:
                libs_versions[lib] = lversion[1]
            else:
                libs_versions[lib] = translate(
                    'VTApp', 'not available',
                    'Part of the library not found text')

        # Construct the dialog
        versions_dlg = QtWidgets.QDialog(self.gui)
        versions_dlg.setWindowTitle(translate('VTApp', 'Version Numbers',
                                              'Caption of the Versions dialog'))
        layout = QtWidgets.QVBoxLayout(versions_dlg)
        versions_edit = QtWidgets.QTextEdit(versions_dlg)
        buttons_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
        layout.addWidget(versions_edit)
        layout.addWidget(buttons_box)

        buttons_box.accepted.connect(versions_dlg.accept)

        versions_edit.setReadOnly(1)
        versions_edit.setText(
            """
            <qt>
            <h3>{title}</h3>
            <table>
            <tr><td><b>Python</b></td><td>{Python}</td></tr>
            <tr><td><b>PyTables</b></td><td>{PyTables}</td></tr>
            <tr><td><b>NumPy</b></td><td>{NumPy}</td></tr>
            <tr><td><b>HDF5</b></td><td>{HDF5}</td></tr>
            <tr><td><b>Zlib</b></td><td>{Zlib}</td></tr>
            <tr><td><b>LZO</b></td><td>{LZO}</td></tr>
            <tr><td><b>BZIP2</b></td><td>{BZIP2}</td></tr>
            <tr><td><b>Qt</b></td><td>{Qt}</td></tr>
            <tr><td><b>PyQt</b></td><td>{PyQt}</td></tr>
            <tr><td><b>ViTables</b></td><td>{ViTables}</td></tr>
            </table>
            </qt>""".format(
                title=libs_versions['title'],
                Python=libs_versions['Python'],
                PyTables=libs_versions['PyTables'],
                NumPy=libs_versions['NumPy'],
                Qt=libs_versions['Qt'],
                PyQt=libs_versions['PyQt'],
                ViTables=libs_versions['ViTables'],
                HDF5=libs_versions['HDF5'],
                Zlib=libs_versions['Zlib'],
                LZO=libs_versions['LZO'],
                BZIP2=libs_versions['BZIP2'],
            ))

        # Show the dialog
        versions_dlg.exec_()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
A form with tables.Group information collected by the
:mod:`vitables.nodeprops.nodeinfo` module.
"""

import os.path

from qtpy import QtGui
from qtpy import QtWidgets

from qtpy.uic import loadUiType

import vitables.utils

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate
# This method of the PyQt5.uic module allows for dinamically loading user
# interfaces created by QtDesigner. See the PyQt5 Reference Guide for more
# info.
Ui_GroupPropPage = \
    loadUiType(os.path.join(os.path.dirname(__file__),'group_prop_page.ui'))[0]



class GroupPropPage(QtWidgets.QWidget, Ui_GroupPropPage):
    """
    Group properties page.

    By loading UI files at runtime we can:

        - create user interfaces at runtime (without using pyuic)
        - use multiple inheritance, MyParentClass(BaseClass, FormClass)

    This class displays a tabbed dialog that shows some properties of
    the selected node. First tab, General, shows general properties like
    name, path, type etc. The second and third tabs show the system and
    user attributes in a tabular way.

    Beware that data types shown in the General page are `PyTables` data
    types so we can deal with `enum`, `time64` and `pseudoatoms` (none of them
    are supported by ``numpy``).
    However data types shown in the System and User attributes pages are
    ``numpy`` data types because `PyTables` attributes are stored as ``numpy``
    arrays.

    :Parameter info: a :meth:`vitables.nodeprops.nodeinfo.NodeInfo` instance
      describing a given node
    """

    def __init__(self, info):
        """Setup the File Properties page."""

        vtapp = vitables.utils.getVTApp()
        super(GroupPropPage, self).__init__(vtapp.gui)
        self.setupUi(self)

        # Customise the dialog's pages
        self.fillGeneralPage(info)


    def fillGeneralPage(self, info):
        """Make the General page of the Properties dialog.

        The page contains two groupboxes that are laid out vertically.

        :Parameter info: a :meth:`vitables.nodeprops.nodeinfo.NodeInfo` instance
          describing a given node
        """

        if info.node_type == 'root group':
            self.nameLE.setText(info.filename)
            self.pathLE.setText(info.filepath)
            self.pathLE.setToolTip(info.filepath)
            self.typeLE.setText(info.file_type)
            self.modeLE.setText(info.mode)
        else:
            self.nameLE.setText(info.nodename)
            self.pathLE.setText(info.nodepath)
            self.pathLE.setToolTip(info.nodepath)
            self.typeLE.setText(info.node_type)

        # Number of children label
        self.nchildrenLE.setText(str(len(info.hanging_nodes)))

        # The group's children table
        table = self.nchildrenTable
        table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        background = table.palette().brush(QtGui.QPalette.Window).color()
        table.setStyleSheet("background-color: {0}".format(background.name()))
        self.children_model = QtGui.QStandardItemModel()
        self.children_model.setHorizontalHeaderLabels([
            translate('GroupPropPage', 'Child name',
            'First column header of the table'),
            translate('GroupPropPage', 'Type',
            'Second column header of the table')])
        table.setModel(self.children_model)
        for name in info.hanging_groups.keys():
            name_item = QtGui.QStandardItem(name)
            type_item = QtGui.QStandardItem(translate('GroupPropPage',
                'group'))
            self.children_model.appendRow([name_item, type_item])
        for name in info.hanging_leaves.keys():
            name_item = QtGui.QStandardItem(name)
            type_item = QtGui.QStandardItem(translate('GroupPropPage', 'leaf'))
            self.children_model.appendRow([name_item, type_item])
        for name in info.hanging_links.keys():
            name_item = QtGui.QStandardItem(name)
            type_item = QtGui.QStandardItem(translate('GroupPropPage', 'link'))
            self.children_model.appendRow([name_item, type_item])


    def regularGroupPage(self):
        """Remove widgets regarding root groups.
        """

        # Remove the Access mode widgets
        self.dblayout.removeWidget(self.modeLabel)
        self.dblayout.removeWidget(self.modeLE)
        self.modeLabel.deleteLater()
        self.modeLE.deleteLater()

        # Change the title of the group box
        self.bottomGB.setTitle(translate('GroupPropPage', 'Group',
            'Title of the group box'))



#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
Setup script for the vitables package.
"""

import os
from setuptools import setup, find_packages


# Utility function to read the README file.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


install_requires = ['qtpy (>=1.2.1)', 'numpy (>=1.4.1)', 'numexpr (>=2.0)',
                    'tables (>=3.0)']
if 'CONDA_PREFIX' not in os.environ:
    install_requires.append('PyQt5 (>=5.5.1)')

setup(name='ViTables',
      version=read('VERSION'),
      description='A viewer for PyTables package',
      long_description=read('README.txt'),
      author='Vicent Mas',
      author_email='vmas@vitables.org',
      maintainer='Vicent Mas',
      maintainer_email='vmas@vitables.org',
      url='https://vitables.org',
      license='GPLv3, see the LICENSE.txt file for detailed info',
      keywords='HDF5 PyTables',
      platforms='Unix, MacOSX, Windows',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: X11 Applications',
          'Environment :: MacOS X',
          'Environment :: Win32 (MS Windows)',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Topic :: Scientific/Engineering'
      ],
      install_requires=install_requires,
      entry_points={
          'console_scripts': ['vitables = vitables.start:gui'],
          'vitables.plugins':
          [('columnar_org = '
            'vitables.plugins.columnorg.columnar_org:ArrayColsOrganizer'),
           ('dbs_tree_sort = '
            'vitables.plugins.dbstreesort.dbs_tree_sort:DBsTreeSort'),
           ('time_series = '
            'vitables.plugins.timeseries.time_series:TSFormatter')]
      },
      packages=find_packages(),
      # Include data files that belong to the vitables package
      package_data={
                'vitables': ['LICENSE.html',
                             'htmldocs/*', 'htmldocs/*/*',
                             'icons/*', 'icons/*/*',
                             '*/*.ui', 'plugins/*/*.ui',
                             'plugins/*/*.ini',
                             'i18n/*.qm'
                             ]
      },
      # In order to include files that do not belong to any package we have to
      # add them to MANIFEST.in AND set the include_package_data keyword to
      # True
      # include_package_data=True,
      )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module provides de `ViTables` GUI: main window, menus, context menus,
toolbars, statusbars and `QActions` bound to both menus and toolbars.
"""

import logging
from vitables.calculator import calculator
import vitables.utils

from qtpy import QtCore
from qtpy import QtGui
from qtpy import QtWidgets

import vitables.h5db.dbstreemodel as dbstreemodel
import vitables.h5db.dbstreeview as dbstreeview
import vitables.logger as logger

__docformat__ = 'restructuredtext'

translate = QtWidgets.QApplication.translate

_GUI_LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


class VTGUI(QtWidgets.QMainWindow):
    """
    The application GUI.

    :Parameters:
        - `vtapp`: an instance of the :meth:`vitables.vtapp.VTApp` class
        - `version`: the `ViTables` version
    """

    def __init__(self, vtapp, version):
        """Initialize the application main window."""

        super(VTGUI, self).__init__(None)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setWindowTitle(translate('VTGUI', 'ViTables {0}',
                                      'Main window title').format(version))
        self.vtapp = vtapp

        # Make the main window easily accessible for external modules
        self.setObjectName('VTGUI')

        self.icons_dictionary = vitables.utils.getIcons()
        # After a node editing the workspace and the databases tree view
        # should not been sync because keyboard focus should remain on
        # the tree view. However, if editing dialogs are raised, when
        # they are closed the focus goes temporarily to the workspace,
        # resulting in spurious syncs. The editing_dlg flag prevents
        # this behavior
        self.editing_dlg = None
        self.logger_dock = None
        self.logger = None
        self.setup_logger_window()
        self.setup(vtapp)

    def setup(self, vtapp):
        """Add widgets, actions, menus and toolbars to the GUI.

        :Parameter tree_view: The databases tree view
        """

        self.dbs_tree_model = dbstreemodel.DBsTreeModel(self, vtapp)
        self.dbs_tree_view = dbstreeview.DBsTreeView(vtapp, self,
                                                     self.dbs_tree_model,
                                                     self)
        self.addComponents()
        self.gui_actions = self.setupActions()
        self.setupToolBars()
        self.setupMenus()
        self.initStatusBar()

        self.logger.nodeCopyAction = self.gui_actions['nodeCopy']

    def setup_logger_window(self):
        # Put the logging console in the bottom region of the window
        self.logger_dock = QtWidgets.QDockWidget('Log')
        self.logger_dock.setObjectName('LoggerDockWidget')
        self.logger_dock.setFeatures(
            QtWidgets.QDockWidget.DockWidgetClosable
            | QtWidgets.QDockWidget.DockWidgetMovable
            | QtWidgets.QDockWidget.DockWidgetFloatable)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.logger_dock)
        self.logger = logger.Logger(self)
        self.logger.setObjectName('LoggerWidget')
        self.logger_dock.setWidget(self.logger)
        # add self.logger as handler of main logger object
        vitables_logger = logging.getLogger('vitables')
        stream_handler = logging.StreamHandler(self.logger)
        stream_handler.setFormatter(logging.Formatter(_GUI_LOG_FORMAT))
        vitables_logger.addHandler(stream_handler)

    def addComponents(self):
        """Add widgets to the main window.

        The main window contains a databases tree view, a workspace and a
        console.

        """

        self.setIconSize(QtCore.QSize(22, 22))
        self.setWindowIcon(self.icons_dictionary['vitables_wm'])
        central_widget = QtWidgets.QWidget(self)
        central_layout = QtWidgets.QVBoxLayout(central_widget)
        # Divide the top region of the window into 2 regions and put there
        # the workspace. The tree of databases will be added later on
        self.hsplitter = QtWidgets.QSplitter(
            QtCore.Qt.Horizontal, central_widget)
        self.hsplitter.setObjectName('hsplitter')
        central_layout.addWidget(self.hsplitter)
        self.setCentralWidget(central_widget)
        self.hsplitter.addWidget(self.dbs_tree_view)
        self.workspace = QtWidgets.QMdiArea(self.hsplitter)
        sb_as_needed = QtCore.Qt.ScrollBarAsNeeded
        self.workspace.setHorizontalScrollBarPolicy(sb_as_needed)
        self.workspace.setVerticalScrollBarPolicy(sb_as_needed)
        self.workspace.setWhatsThis(
            translate('VTGUI',
                      """<qt>
                      <h3>The Workspace</h3>
                      This is the area where open leaves of the object tree are
                      displayed. Many tables and arrays can be displayed
                      simultaneously.
                      <p>The diferent views can be tiled as a mosaic or
                      stacked as a cascade.
                      </qt>""",
                      'WhatsThis help for the workspace'))

        # The signal mapper used to keep the the Window menu updated
        self.window_mapper = QtCore.QSignalMapper(self)
        self.window_mapper.mapped[QtWidgets.QWidget].connect(
            self.workspace.setActiveSubWindow)
        self.workspace.installEventFilter(self)
        self.dbs_tree_view.clicked.connect(self.selection_changed)

    def selection_changed(self, index):
        self.updateActions()
        self.updateStatusBar()

    def setupActions(self):
        """Provide actions to the menubar and the toolbars.
        """

        # Setting action names makes it easier to acces these actions
        # from plugins
        actions = {}
        actions['fileNew'] = QtWidgets.QAction(
            translate('VTGUI', '&New...', 'File -> New'), self,
            shortcut=QtGui.QKeySequence.New,
            triggered=self.vtapp.fileNew,
            icon=self.icons_dictionary['document-new'],
            statusTip=translate(
                'VTGUI', 'Create a new file',
                'Status bar text for the File -> New action'))
        actions['fileNew'].setObjectName('fileNew')

        actions['fileOpen'] = QtWidgets.QAction(
            translate('VTGUI', '&Open...', 'File -> Open'), self,
            shortcut=QtGui.QKeySequence.Open,
            triggered=self.vtapp.fileOpen,
            icon=self.icons_dictionary['document-open'],
            statusTip=translate(
                'VTGUI', 'Open an existing file',
                'Status bar text for the File -> Open action'))
        actions['fileOpen'].setObjectName('fileOpen')

        actions['fileOpenRO'] = QtWidgets.QAction(
            translate('VTGUI', 'Read-only open...', 'File -> Open'), self,
            triggered=self.vtapp.fileOpenRO,
            icon=self.icons_dictionary['document-open'],
            statusTip=translate(
                'VTGUI', 'Open an existing file in read-only mode',
                'Status bar text for the File -> Open action'))
        actions['fileOpenRO'].setObjectName('fileOpenRO')

        actions['fileClose'] = QtWidgets.QAction(
            translate('VTGUI', '&Close', 'File -> Close'), self,
            shortcut=QtGui.QKeySequence('Shift+F4'),
            triggered=self.vtapp.fileClose,
            icon=self.icons_dictionary['document-close'],
            statusTip=translate(
                'VTGUI', 'Close the selected file',
                'Status bar text for the File -> Close action'))
        actions['fileClose'].setObjectName('fileClose')

        actions['fileCloseAll'] = QtWidgets.QAction(
            translate('VTGUI', 'Close &All', 'File -> Close All'), self,
            triggered=self.vtapp.fileCloseAll,
            statusTip=translate(
                'VTGUI', 'Close all files',
                'Status bar text for the File -> Close All action'))
        actions['fileCloseAll'].setObjectName('fileCloseAll')

        actions['fileSaveAs'] = QtWidgets.QAction(
            translate('VTGUI', '&Save as...', 'File -> Save As'), self,
            shortcut=QtGui.QKeySequence.SaveAs,
            triggered=self.vtapp.fileSaveAs,
            icon=self.icons_dictionary['document-save-as'],
            statusTip=translate(
                'VTGUI', 'Save a renamed copy of the selected file',
                'Status bar text for the File -> Save As action'))
        actions['fileSaveAs'].setObjectName('fileSaveAs')

        actions['fileExit'] = QtWidgets.QAction(
            translate('VTGUI', 'E&xit', 'File -> Exit'), self,
            shortcut=QtGui.QKeySequence.Quit,
            triggered=self.close,
            icon=self.icons_dictionary['application-exit'],
            statusTip=translate(
                'VTGUI', 'Quit ViTables',
                'Status bar text for the File -> Exit action'))
        actions['fileExit'].setObjectName('fileExit')

        actions['nodeOpen'] = QtWidgets.QAction(
            translate('VTGUI', '&Open view', 'Node -> Open View'), self,
            shortcut=QtGui.QKeySequence('Alt+Ctrl+O'),
            triggered=self.vtapp.nodeOpen,
            statusTip=translate(
                'VTGUI', 'Display the contents of the selected node',
                'Status bar text for the Node -> Open View action'))
        actions['nodeOpen'].setObjectName('nodeOpen')

        actions['nodeClose'] = QtWidgets.QAction(
            translate('VTGUI', 'C&lose view', 'Node -> Close View'), self,
            shortcut=QtGui.QKeySequence('Alt+Shift+F4'),
            triggered=self.vtapp.nodeClose,
            statusTip=translate(
                'VTGUI', 'Close the view of the selected node',
                'Status bar text for the Node -> Close View action'))
        actions['nodeClose'].setObjectName('nodeClose')

        actions['nodeProperties'] = QtWidgets.QAction(
            translate('VTGUI', 'Prop&erties...', 'Node -> Properties'), self,
            shortcut=QtGui.QKeySequence('Ctrl+I'),
            triggered=self.vtapp.nodeProperties,
            icon=self.icons_dictionary['help-about'],
            statusTip=translate(
                'VTGUI', 'Show the properties dialog for the selected node',
                'Status bar text for the Node -> Properties action'))
        actions['nodeProperties'].setObjectName('nodeProperties')

        actions['nodeNew'] = QtWidgets.QAction(
            translate('VTGUI', '&New group...', 'Node -> New group'), self,
            shortcut=QtGui.QKeySequence('Alt+Ctrl+N'),
            triggered=self.vtapp.nodeNewGroup,
            icon=self.icons_dictionary['folder-new'],
            statusTip=translate(
                'VTGUI', 'Create a new group under the selected node',
                'Status bar text for the Node -> New group action'))
        actions['nodeNew'].setObjectName('nodeNew')

        actions['nodeRename'] = QtWidgets.QAction(
            translate('VTGUI', '&Rename...', 'Node -> Rename'), self,
            shortcut=QtGui.QKeySequence('Ctrl+R'),
            triggered=self.vtapp.nodeRename,
            icon=self.icons_dictionary['edit-rename'],
            statusTip=translate(
                'VTGUI', 'Rename the selected node',
                'Status bar text for the Node -> Rename action'))
        actions['nodeRename'].setObjectName('nodeRename')

        actions['nodeCut'] = QtWidgets.QAction(
            translate('VTGUI', 'Cu&t', 'Node -> Cut'), self,
            shortcut=QtGui.QKeySequence.Cut,
            triggered=self.vtapp.nodeCut,
            icon=self.icons_dictionary['edit-cut'],
            statusTip=translate('VTGUI', 'Cut the selected node',
                                'Status bar text for the Node -> Cut action'))
        actions['nodeCut'].setObjectName('nodeCut')

        actions['nodeCopy'] = QtWidgets.QAction(
            translate('VTGUI', '&Copy', 'Node -> Copy'), self,
            shortcut=QtGui.QKeySequence.Copy,
            triggered=self.makeCopy,
            icon=self.icons_dictionary['edit-copy'],
            statusTip=translate('VTGUI', 'Copy the selected node',
                                'Status bar text for the Node -> Copy action'))
        actions['nodeCopy'].setObjectName('nodeCopy')

        actions['nodePaste'] = QtWidgets.QAction(
            translate('VTGUI', '&Paste', 'Node -> Paste'), self,
            shortcut=QtGui.QKeySequence.Paste,
            triggered=self.vtapp.nodePaste,
            icon=self.icons_dictionary['edit-paste'],
            statusTip=translate('VTGUI', 'Paste the last copied/cut node',
                                'Status bar text for the Node -> Copy action'))
        actions['nodePaste'].setObjectName('nodePaste')

        actions['nodeDelete'] = QtWidgets.QAction(
            translate('VTGUI', '&Delete', 'Node -> Delete'), self,
            shortcut=QtGui.QKeySequence.Delete,
            triggered=self.vtapp.nodeDelete,
            icon=self.icons_dictionary['edit-delete'],
            statusTip=translate('VTGUI', 'Delete the selected node',
                                'Status bar text for the Node -> Copy action'))
        actions['nodeDelete'].setObjectName('nodeDelete')

        actions['queryNew'] = QtWidgets.QAction(
            translate('VTGUI', '&Query...', 'Query -> New...'), self,
            triggered=self.vtapp.newQuery,
            icon=self.icons_dictionary['view-filter'],
            statusTip=translate(
                'VTGUI', 'Create a new filter for the selected table',
                'Status bar text for the Query -> New... action'))
        actions['queryNew'].setObjectName('queryNew')

        actions['queryDeleteAll'] = QtWidgets.QAction(
            translate('VTGUI', 'Delete &All', 'Query -> Delete All'), self,
            triggered=self.vtapp.deleteAllQueries,
            icon=self.icons_dictionary['delete_filters'],
            statusTip=translate(
                'VTGUI', 'Remove all filters',
                'Status bar text for the Query -> Delete All action'))
        actions['queryDeleteAll'].setObjectName('queryDeleteAll')

        actions['settingsPreferences'] = QtWidgets.QAction(
            translate('VTGUI', '&Preferences...', 'Settings -> Preferences'),
            self,
            shortcut=QtGui.QKeySequence.Preferences,
            triggered=self.vtapp.settingsPreferences,
            icon=self.icons_dictionary['configure'],
            statusTip=translate(
                'VTGUI', 'Configure ViTables',
                'Status bar text for the Settings -> Preferences action'))
        actions['settingsPreferences'].setObjectName('settingsPreferences')

        actions['windowCascade'] = QtWidgets.QAction(
            translate('VTGUI', '&Cascade', 'Windows -> Cascade'), self,
            triggered=self.workspace.cascadeSubWindows,
            statusTip=translate(
                'VTGUI', 'Arranges open windows in a cascade pattern',
                'Status bar text for the Windows -> Cascade action'))
        actions['windowCascade'].setObjectName('windowCascade')

        actions['windowTile'] = QtWidgets.QAction(
            translate('VTGUI', '&Tile', 'Windows -> Tile'), self,
            triggered=self.workspace.tileSubWindows,
            statusTip=translate(
                'VTGUI', 'Arranges open windows in a tile pattern',
                'Status bar text for the Windows -> Tile action'))
        actions['windowTile'].setObjectName('windowTile')

        actions['windowRestoreAll'] = QtWidgets.QAction(
            translate('VTGUI', '&Restore All', 'Windows -> Restore All'),
            self,
            triggered=self.vtapp.windowRestoreAll,
            statusTip=translate(
                'VTGUI', 'Restore all minimized windows on the workspace',
                'Status bar text for the Windows -> Restore All action'))
        actions['windowRestoreAll'].setObjectName('windowRestoreAll')

        actions['windowMinimizeAll'] = QtWidgets.QAction(
            translate('VTGUI', '&Minimize All', 'Windows -> Minimize All'),
            self,
            triggered=self.vtapp.windowMinimizeAll,
            statusTip=translate(
                'VTGUI', 'Minimize all windows on the workspace',
                'Status bar text for the Windows -> Restore All action'))
        actions['windowMinimizeAll'].setObjectName('windowMinimizeAll')

        actions['windowClose'] = QtWidgets.QAction(
            translate('VTGUI', 'C&lose', 'Windows -> Close'), self,
            triggered=self.vtapp.windowClose,
            statusTip=translate(
                'VTGUI', 'Close the active view',
                'Status bar text for the Windows -> Close action'))
        actions['windowClose'].setObjectName('windowClose')

        actions['windowCloseAll'] = QtWidgets.QAction(
            translate('VTGUI', 'Close &All', 'Windows -> Close All'), self,
            triggered=self.vtapp.windowCloseAll,
            statusTip=translate(
                'VTGUI', 'Close all views',
                'Status bar text for the Windows -> Close All action'))
        actions['windowCloseAll'].setObjectName('windowCloseAll')

        actions['windowSeparator'] = QtWidgets.QAction(
            translate('VTGUI', 'Current View', 'Windows -> separator'), self)
        actions['windowSeparator'].setSeparator(True)
        actions['windowSeparator'].setObjectName('windowSeparator')

        actions['mdiTabbed'] = QtWidgets.QAction(
            translate('VTGUI', 'Toggle MDI/Tabs', 'MDI -> Tabbed'), self,
            triggered=self.changeMDIViewMode,
            statusTip=translate(
                'VTGUI',
                'Toggle workspace view mode between Tabs <--> Multi Document '
                'Interface',
                'Status bar text for the MDI -> Tabbed action'))
        actions['mdiTabbed'].setObjectName('mdiTabbed')

        actions['helpUsersGuide'] = QtWidgets.QAction(
            translate('VTGUI', "&User's Guide", 'Help -> Users Guide'), self,
            shortcut=QtGui.QKeySequence.HelpContents,
            triggered=self.vtapp.helpBrowser,
            icon=self.icons_dictionary['help-contents'],
            statusTip=translate(
                'VTGUI', "Open the ViTables User's Guide",
                'Status bar text for the Help -> Users Guide action'))
        actions['helpUsersGuide'].setObjectName('helpUsersGuide')

        actions['helpAbout'] = QtWidgets.QAction(
            translate('VTGUI', '&About ViTables', 'Help -> About'), self,
            triggered=self.vtapp.helpAbout,
            icon=self.icons_dictionary['vitables_wm'],
            statusTip=translate(
                'VTGUI', 'Display information about ViTables',
                'Status bar text for the Help -> About action'))
        actions['helpAbout'].setObjectName('helpAbout')

        actions['helpAboutQt'] = QtWidgets.QAction(
            translate('VTGUI', 'About &Qt', 'Help -> About Qt'), self,
            triggered=self.vtapp.helpAboutQt,
            statusTip=translate(
                'VTGUI', 'Display information about the Qt library',
                'Status bar text for the Help -> About Qt action'))
        actions['helpAboutQt'].setObjectName('helpAboutQt')

        actions['helpVersions'] = QtWidgets.QAction(
            translate('VTGUI', 'Show &Versions', 'Help -> Show Versions'),
            self,
            triggered=self.vtapp.helpVersions,
            statusTip=translate(
                'VTGUI', 'Show the versions of the libraries used by ViTables',
                'Status bar text for the Help -> Show Versions action'))
        actions['helpVersions'].setObjectName('helpVersions')

        actions['calculate'] = QtWidgets.QAction(
            translate('VTGUI', 'Calculate...', 'Calculate action'),
            self,
            triggered=calculator.run,
            statusTip=translate(
                'VTGUI', 'Run calculation on opened tables.',
                'Action tip'))
        actions['calculate'].setObjectName('calculate')

        return actions

    def setupToolBars(self):
        """
        Set up the main window toolbars.

        Toolbars are made of actions.
        """

        # File toolbar
        self.file_toolbar = self.addToolBar(
            translate('VTGUI', 'File operations', 'Toolbar title'))
        # Warning! Do NOT use 'File toolbar' as a object name or it will
        # show an strange behaviour (a Qt bug I think): it will always
        # be added to the left and will expand the whole top area
        self.file_toolbar.setObjectName('File toolbar')
        actions = ['fileNew', 'fileOpen', 'fileClose', 'fileSaveAs']
        vitables.utils.addActions(self.file_toolbar, actions, self.gui_actions)

        # Reset the tooltip of the File -> Open... button
        file_open_button = self.file_toolbar.widgetForAction(
            self.gui_actions['fileOpen'])
        file_open_button.setToolTip(translate(
            'VTGUI', """Click to open a """
            """file\nClick and hold to open a recent file""",
            'File toolbar -> Open Recent Files'))

        # Node toolbar
        self.node_toolbar = self.addToolBar(
            translate('VTGUI', 'Node operations', 'Toolbar title'))
        self.node_toolbar.setObjectName('Node toolbar')
        actions = ['nodeNew', 'nodeCut', 'nodeCopy', 'nodePaste', 'nodeDelete']
        vitables.utils.addActions(self.node_toolbar, actions, self.gui_actions)

        # Query toolbar
        self.query_toolbar = self.addToolBar(
            translate('VTGUI', 'Queries on tables', 'Toolbar title'))
        self.query_toolbar.setObjectName('Query toolbar')
        actions = ['queryNew', 'queryDeleteAll']
        vitables.utils.addActions(self.query_toolbar, actions,
                                  self.gui_actions)

        # Help toolbar
        self.help_toolbar = self.addToolBar(translate('VTGUI', 'Help system',
                                                      'Toolbar title'))
        self.help_toolbar.setObjectName('Help toolbar')
        actions = ['helpUsersGuide']
        vitables.utils.addActions(self.help_toolbar, actions, self.gui_actions)
        whatis = QtWidgets.QWhatsThis.createAction(self.help_toolbar)
        whatis.setObjectName('whatis_help_toolbar')
        whatis.setStatusTip(translate(
            'VTGUI', 'Whats this? help for a widget',
            'Status bar text for the Help -> Whats This action'))
        whatis.setShortcut(QtGui.QKeySequence())
        self.help_toolbar.addAction(whatis)

    def initStatusBar(self):
        """Initialise the status bar."""

        status_bar = self.statusBar()
        self.sb_node_info = QtWidgets.QLabel(status_bar)
        self.sb_node_info.setObjectName('status bar widget')
        self.sb_node_info.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding,
                                        QtWidgets.QSizePolicy.Minimum)
        status_bar.addPermanentWidget(self.sb_node_info)
        self.sb_node_info.setToolTip(translate(
            'VTGUI', 'The node currently selected in the Tree of '
            'databases pane', 'The Selected node box startup message'))
        status_bar.showMessage(translate('VTGUI', 'Ready...',
                                         'The status bar startup message'))

    def setupMenus(self):
        """
        Set up the main window menus.

        Popups are made of actions, items and separators.
        The `Window` menu is a special case due to its dynamic nature. Its
        contents depend on the number of existing views.
        In order to track changes and keep updated the menu, it is reloaded
        every time it is about to be displayed.
        """

        # Create the File menu and add actions/submenus/separators to it
        self.file_menu = self.menuBar().addMenu(translate(
            'VTGUI', "&File", 'The File menu entry'))
        self.file_menu.setObjectName('file_menu')
        # The Open Recent submenu
        self.open_recent_submenu = QtWidgets.QMenu(
            translate('VTGUI', 'Open R&ecent Files',
                      'File -> Open Recent Files'))
        self.open_recent_submenu.setObjectName('open_recent_submenu')
        self.open_recent_submenu.setSeparatorsCollapsible(False)
        self.open_recent_submenu.setIcon(
            self.icons_dictionary['document-open-recent'])

        file_actions = ['fileNew', 'fileOpen', 'fileOpenRO',
                        self.open_recent_submenu, None, 'fileClose',
                        'fileCloseAll', None, 'fileSaveAs', None, 'fileExit']
        vitables.utils.addActions(self.file_menu, file_actions,
                                  self.gui_actions)

        file_open_button = self.file_toolbar.widgetForAction(
            self.gui_actions['fileOpen'])
        file_open_button.setMenu(self.open_recent_submenu)
        self.open_recent_submenu.aboutToShow.connect(self.updateRecentSubmenu)

        # Create the Node menu and add actions/submenus/separators to it
        self.node_menu = self.menuBar().addMenu(
            translate('VTGUI', "&Node", 'The Node menu entry'))
        self.node_menu.setObjectName('node_menu')
        node_actions = ['nodeOpen', 'nodeClose', 'nodeProperties', None,
                        'nodeNew', 'nodeRename', 'nodeCut', 'nodeCopy',
                        'nodePaste', 'nodeDelete']
        vitables.utils.addActions(self.node_menu, node_actions,
                                  self.gui_actions)

        # Create the Dataset menu and add actions/submenus/separators to it
        self.dataset_menu = self.menuBar().addMenu(
            translate('VTGUI', "&Dataset", 'The Dataset menu entry'))
        self.dataset_menu.setObjectName('dataset_menu')
        dataset_actions = ['queryNew', 'calculate']
        vitables.utils.addActions(self.dataset_menu, dataset_actions,
                                  self.gui_actions)

        # Create the Settings menu and add actions/submenus/separators to it
        settings_menu = self.menuBar().addMenu(
            translate('VTGUI', "&Settings", 'The Settings menu entry'))
        settings_menu.setObjectName('settings_menu')
        # Returns a popup menu containing checkable entries for the toolbars
        # and dock widgets present in the main window
        self.toolbars_submenu = self.createPopupMenu()
        self.toolbars_submenu.setObjectName('settings_toolbars_submenu')
        self.toolbars_submenu.menuAction().setText(
            translate('VTGUI', 'ToolBars', 'Tools -> ToolBars action'))
        settings_actions = ['settingsPreferences', None,
                            self.toolbars_submenu]
        vitables.utils.addActions(settings_menu, settings_actions,
                                  self.gui_actions)

        # Create the Window menu and add actions/menus/separators to it
        self.window_menu = self.menuBar().addMenu(
            translate('VTGUI', "&Window", 'The Windows menu entry'))
        self.window_menu.setObjectName('window_menu')
        windows_actions = ['windowCascade', 'windowTile',
                           'windowRestoreAll', 'windowMinimizeAll',
                           'windowClose', 'windowCloseAll', None, 'mdiTabbed',
                           'windowSeparator']
        vitables.utils.addActions(self.window_menu, windows_actions,
                                  self.gui_actions)
        self.window_menu.setSeparatorsCollapsible(True)
        self.window_menu.aboutToShow.connect(self.updateWindowMenu)

        # Create the Help menu and add actions/menus/separators to it
        help_menu = self.menuBar().addMenu(
            translate('VTGUI', "&Help", 'The Help menu entry'))
        help_menu.setObjectName('help_menu')
        help_actions = ['helpUsersGuide', None, 'helpAbout', 'helpAboutQt',
                        'helpVersions', None]
        vitables.utils.addActions(help_menu, help_actions, self.gui_actions)
        whatis = QtWidgets.QWhatsThis.createAction(help_menu)
        whatis.setObjectName('whatis_help_menu')
        whatis.setStatusTip(
            translate('VTGUI', 'Context help',
                      'Status bar text for the Help -> Whats This action'))
        help_menu.addAction(whatis)

        #########################################################
        #
        # 				Context menus
        #
        #########################################################

        self.view_cm = QtWidgets.QMenu(self)
        self.view_cm.setObjectName('view_cm')
        actions = ['fileNew', 'fileOpen', 'fileOpenRO',
                   self.open_recent_submenu, None, 'fileClose',
                   'fileCloseAll', None, 'fileSaveAs', None, 'fileExit']
        vitables.utils.addActions(self.view_cm, actions, self.gui_actions)

        self.root_node_cm = QtWidgets.QMenu(self)
        self.root_node_cm.setObjectName('root_node_cm')
        actions = ['fileClose', 'fileSaveAs', None, 'nodeProperties', None,
                   'nodeNew', 'nodeCopy', 'nodePaste', None, 'queryDeleteAll']
        vitables.utils.addActions(self.root_node_cm, actions, self.gui_actions)

        self.group_node_cm = QtWidgets.QMenu(self)
        self.group_node_cm.setObjectName('group_node_cm')
        actions = ['nodeProperties', None, 'nodeNew', 'nodeRename', 'nodeCut',
                   'nodeCopy', 'nodePaste', 'nodeDelete']
        vitables.utils.addActions(self.group_node_cm, actions,
                                  self.gui_actions)

        self.leaf_node_cm = QtWidgets.QMenu(self)
        self.leaf_node_cm.setObjectName('leaf_node_cm')
        actions = ['nodeOpen', 'nodeClose', None, 'nodeProperties', None,
                   'nodeRename', 'nodeCut', 'nodeCopy', 'nodePaste',
                   'nodeDelete', None, 'queryNew']
        vitables.utils.addActions(self.leaf_node_cm, actions, self.gui_actions)

        self.mdi_cm = QtWidgets.QMenu(self)
        self.mdi_cm.setObjectName('mdi_cm')
        actions = [self.window_menu]
        vitables.utils.addActions(self.mdi_cm, actions, self.gui_actions)

    def closeEvent(self, event):
        """
        Handle close events.

        Clicking the close button of the main window titlebar causes
        the application quitting immediately, leaving things in a non
        consistent state. This event handler ensures that the needed
        tidy up is done before quitting.

        :Parameter event: the close event being handled
        """

        # Do the required tidy up
        self.vtapp.fileExit()

        # Quit
        event.accept()
        QtWidgets.qApp.quit()

    def makeCopy(self):
        """Copy text/leaf depending on which widget has focus.

        This method disambiguates the ``Ctrl+C`` shortcut. If the console has
        focus then ``Ctrl+C`` will copy the console selected text. If the
        databases tree view has focus then the selected node (if any) will be
        copied.
        """

        if self.dbs_tree_view.hasFocus():
            self.vtapp.nodeCopy()
        elif self.logger.hasFocus():
            self.logger.copy()

    # Updating appearance means:
    #
    #     * changing the toolbar buttons look when their tied QActions are
    #       enabled/disabled
    #     * updating content of menus and submenus
    #
    # Updating state means:
    #
    #     * toggling state of QActions i.e. enabling/disabling QActions
    #

    def updateActions(self):
        """
        Update the state of the actions tied to menu items and toolbars.

        Every time that the selected item changes in the tree viewer the
        state of the actions must be updated because it depends on the
        type of selected item (leaf or group, opening mode etc.).
        The following events trigger a call to this slot:

            * insertion/deletion of rows in the tree of databases model
            * changes in the selection state of the tree of databases view

        The slot should be manually called when a new view is activated in
        the workspace (for instance by methods
        :meth:`vitables.vtapp.VTApp.nodeOpen`,
        :meth:`vitables.vtapp.VTApp.nodeClose`).

        .. _Warning:

        Warning! Don\'t call this method until the GUI initialisation finishes.
        """

        # The following actions are always active:
        # fileNew, fileOpen, fileOpenRO, fileExit and the Help menu actions

        # The set of actions that can be enabled or disabled
        actions = frozenset(['fileClose', 'fileCloseAll', 'fileSaveAs',
                             'nodeOpen', 'nodeClose', 'nodeProperties',
                             'nodeNew', 'nodeRename', 'nodeCut', 'nodeCopy',
                             'nodePaste', 'nodeDelete',
                             'queryNew', 'queryDeleteAll'])
        enabled = set([])

        model_rows = self.dbs_tree_model.rowCount(QtCore.QModelIndex())
        if model_rows <= 0:
            return

        # If there are open files aside the temporary DB
        if model_rows > 1:
            enabled = enabled.union(['fileCloseAll'])

        # if there are filtered tables --> queryDeleteAll is enabled
        tmp_index = self.dbs_tree_model.index(model_rows - 1, 0,
                                              QtCore.QModelIndex())
        ftables = self.dbs_tree_model.rowCount(tmp_index)
        if ftables > 0:
            enabled = enabled.union(['queryDeleteAll'])

        current = self.dbs_tree_view.currentIndex()
        node = self.dbs_tree_model.nodeFromIndex(current)
        if node != self.dbs_tree_model.root:
            # Actions always enabled for every node
            enabled = enabled.union(['nodeProperties',
                                     'nodeCopy'])

            # If the selected file is not the temporary DB
            if node.filepath != self.dbs_tree_model.tmp_filepath:
                enabled = enabled.union(['fileSaveAs', 'fileClose'])

            kind = node.node_kind
            # If the node is a table --> queryNew is enabled
            if kind == 'table':
                enabled = enabled.union(['queryNew'])

            # If the file is not open in read-only mode
            mode = self.dbs_tree_model.getDBDoc(node.filepath).mode
            if mode != 'r':
                if kind == 'root group':
                    enabled = enabled.union(['nodeNew', 'nodePaste'])
                elif kind == 'group':
                    enabled = enabled.union(['nodeNew', 'nodeRename',
                                             'nodeCut', 'nodePaste',
                                             'nodeDelete'])
                elif kind == 'table':
                    enabled = enabled.union(['nodeRename', 'nodeCut',
                                             'nodeDelete'])
                else:
                    enabled = enabled.union(['nodeRename', 'nodeCut',
                                             'nodeDelete'])

            if kind not in ('group', 'root group'):
                if node.has_view:
                    enabled = enabled.union(['nodeClose'])
                else:
                    enabled = enabled.union(['nodeOpen'])

        disabled = actions.difference(enabled)
        for action in enabled:
            self.gui_actions[action].setEnabled(True)
        for action in disabled:
            self.gui_actions[action].setDisabled(True)

    def updateRecentSubmenu(self):
        """Update the content of the `Open Recent File...` submenu."""

        index = 0
        self.open_recent_submenu.clear()
        iconset = vitables.utils.getIcons()
        for item in self.vtapp.config.recent_files:
            index += 1
            (mode, filepath) = item.split('#@#')
            action = QtWidgets.QAction('{0:>2} {1}.'.format(index, filepath),
                                       self,
                                       triggered=self.vtapp.openRecentFile)
            action.setData(item)
            if mode == 'r':
                action.setIcon(iconset['file_ro'])
            else:
                action.setIcon(iconset['file_rw'])
            self.open_recent_submenu.addAction(action)

        # Always add a separator and a clear QAction. So if the menu is empty
        # the user still will know what's going on
        self.open_recent_submenu.addSeparator()
        action = QtWidgets.QAction(
            translate('VTGUI', '&Clear', 'Recent File submenu entry'),
            self, triggered=self.vtapp.clearRecentFiles)
        self.open_recent_submenu.addAction(action)

    def updateWindowMenu(self):
        """
        Update the `Window` menu.

        The `Window` menu is dynamic because its content is determined
        by the currently open views. Because the number of these views or
        its contents may vary at any moment we must update the `Window`
        menu every time it is open. For simplicity we don't keep track
        of changes in the menu content. Instead, we clean and create it
        from scratch every time it is about to show.
        """

        # Clear the menu and rebuild it from scratch
        wmenu = self.window_menu
        wmenu.clear()
        window_actions = ['windowCascade', 'windowTile', 'windowRestoreAll',
                          'windowMinimizeAll', 'windowClose', 'windowCloseAll',
                          None, 'mdiTabbed', 'windowSeparator']
        vitables.utils.addActions(wmenu, window_actions, self.gui_actions)

        windows_list = self.workspace.subWindowList()
        if not windows_list:
            return

        # Create actions for the existing views
        action_group = QtWidgets.QActionGroup(wmenu)
        wmenu.action_group = action_group
        counter = 1
        for window in windows_list:
            title = window.windowTitle()
            if counter < 10:
                action = wmenu.addAction("&{0:d} {1}".format(counter, title))
                wmenu.action_group.addAction(action)
            elif counter == 10:
                wmenu.addSeparator()
                submenu = wmenu.addMenu(
                    translate('VTGUI', "&More...", 'A Windows submenu'))
                action_subgroup = QtWidgets.QActionGroup(submenu)
            elif counter < 36:
                atext = "&{0} {1}".format(chr(counter + ord("@") - 10), title)
                action = submenu.addAction(atext)
                action_subgroup.addAction(action)

            self.window_mapper.setMapping(action, window)
            action.triggered.connect(self.window_mapper.map)
            action.setCheckable(True)
            if self.workspace.activeSubWindow() == window:
                action.setChecked(True)
            counter = counter + 1

    def updateStatusBar(self):
        """Update the permanent message of the status bar.
        """

        current = self.dbs_tree_view.currentIndex()
        if current.isValid():
            tip = self.dbs_tree_model.data(current, QtCore.Qt.StatusTipRole)
            message = tip
        else:
            message = ''
        self.sb_node_info.setText(message)

    def popupContextMenu(self, kind, pos):
        """
        Popup a context menu in the tree of databases view.

        When a point of the tree view is right clicked, a context
        popup is displayed. The content of the popup depends on the
        kind of node pointed: no node, root group, group or leaf.

        :Parameters:

            - `kind`: defines the content of the menu
            - `pos`: the clicked point in global coordinates
        """

        if kind == 'view':
            menu = self.view_cm
        elif kind == 'root group':
            menu = self.root_node_cm
        elif kind == 'group':
            menu = self.group_node_cm
        else:
            menu = self.leaf_node_cm
        menu.popup(pos)

    def closeChildrenViews(self, nodepath, filepath):
        """Close views being overwritten during node editing.

        :Parameters:

            - `nodepath`: the full path of the node that is overwrting other
                nodes
            - `filepath`: the full path of the file where that node lives
        """

        for window in self.workspace.subWindowList():
            wnodepath = window.dbt_leaf.nodepath
            wfilepath = window.dbt_leaf.filepath
            if not wfilepath == filepath:
                continue
            if wnodepath[0:len(nodepath)] == nodepath:
                window.close()

    def changeMDIViewMode(self):
        """Toggle the view mode of the workspace.
        """

        if self.workspace.viewMode() == QtWidgets.QMdiArea.SubWindowView:
            self.workspace.setViewMode(QtWidgets.QMdiArea.TabbedView)
        else:
            self.workspace.setViewMode(QtWidgets.QMdiArea.SubWindowView)

    def eventFilter(self, widget, event):
        """Event filter used to provide the workspace with a context menu.

        :Parameters:

            - `widget`: the widget that receives the event
            - `event`: the event being processed
        """

        if widget == self.workspace:
            if event.type() == QtCore.QEvent.ContextMenu:
                # If active mdi subwindow can handle context menu
                # pass the event to it
                active_window = self.workspace.activeSubWindow()
                if active_window is not None and \
                        hasattr(active_window, 'is_context_menu_custom'):
                    is_cm_custom = active_window.is_context_menu_custom
                else:
                    is_cm_custom = False
                if not is_cm_custom:
                    pos = event.globalPos()
                    self.mdi_cm.popup(pos)
            return QtWidgets.QMdiArea.eventFilter(widget, widget, event)
        else:
            return QtWidgets.QMainWindow.eventFilter(self, widget, event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""Table with time fields."""

import os
import time
import numpy

import tables

# Describe a particle record
class Particle(tables.IsDescription):
    """This class defines a table record.
    """

    lati        = tables.IntCol(pos=0)
#    longi       = IntCol(pos=1)
    Time        = tables.Time64Col(pos=2)
    pressure    = tables.FloatCol(pos=3)
    ID          = tables.StringCol(itemsize=10, pos=4)
    Int16       = tables.UIntCol(itemsize=4, pos=5)
    Int64       = tables.IntCol(itemsize=8, pos=6)
    Bool        = tables.BoolCol(pos=7)

output_dir = '../timeseries'
try:
    os.mkdir(output_dir)
except OSError:
    pass

# Open a new empty HDF5 file
hdf5_name = "table_ts.h5"
filepath_hdf5 = os.path.join(output_dir, hdf5_name)
h5file = tables.open_file(filepath_hdf5, mode="w",
title='Example Table with time series')

# Get the HDF5 root group
root = h5file.root
group = h5file.create_group(root, "Particles")
filters = tables.Filters(complevel=1, complib='lzo', shuffle=1)
nrows = 6000
table = h5file.create_table("/Particles", "TParticle", Particle,
                          "Sample set of particles ", filters,
                          expectedrows = nrows)

# Number of rows in buffer
nrowsbuf = 1000
# Fill the table with 10**N particles
for i in numpy.arange(0, nrows, nrowsbuf, dtype=numpy.int64):
    if i+nrowsbuf > nrows:
        nrapp = nrows-i
    else:
        nrapp = nrowsbuf
    # First, assign the values to the Particle record
    Int64 = numpy.arange(i, i+nrapp)
    Time = Int64 + time.time()
    lati = Int64
    pressure = lati - 10.4
    ID = (i + Int64).astype('S10')
    Int16 = lati - 50
    Bool = lati % 3 > 1
    # This injects the Record values
    table.append([lati, Time, pressure, ID, Int16, Int64, Bool])

# Flush the table buffers and close the file
h5file.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#       Copyright (C) 2005-2007 Carabos Coop. V. All rights reserved
#       Copyright (C) 2008-2019 Vicent Mas. All rights reserved
#
#       This program is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#       Author:  Vicent Mas - vmas@vitables.org

"""
This module defines an Exception to be used when there is a problem
reading/writting the ``ViTables`` configuration.
"""

__docformat__ = 'restructuredtext'

from qtpy import QtGui
from qtpy import QtWidgets

translate = QtWidgets.QApplication.translate


class ConfigFileIOException(Exception):
    """Exception class for IO errors in the configuration file.

    :Parameter key:
        the configuration file key that cannot be read/written
    """

    def __init__(self, key):
        """Setup the configuration error messages that will be shown to user.
        """

        Exception.__init__(self)
        # If key looks like /path/to/property=value a write exception is
        # raised. If not a read exception is raised
        if '=' in key:
            setting = key.split('=')[0]
            self.error_message = translate('ConfigFileIOException', \
                """\nConfiguration error: the application setting """\
                """{0} cannot be saved.""",
                'A logger error message').format(setting)
        else:
            self.error_message = translate('ConfigFileIOException', \
                """\nConfiguration warning: the application setting """\
                """{0} cannot be read. Its default value will be used.""",
                'A logger error message').format(key)
